from typing import Any
from typing import Dict

from .aurora import nflx_aurora_capacity_model
from .cassandra import nflx_cassandra_capacity_model
from .control import nflx_control_capacity_model
from .counter import nflx_counter_capacity_model
from .crdb import nflx_cockroachdb_capacity_model
from .ddb import nflx_ddb_capacity_model
from .elasticsearch import nflx_elasticsearch_capacity_model
from .elasticsearch import nflx_elasticsearch_data_capacity_model
from .elasticsearch import nflx_elasticsearch_master_capacity_model
from .elasticsearch import nflx_elasticsearch_search_capacity_model
from .entity import nflx_entity_capacity_model
from .evcache import nflx_evcache_capacity_model
from .graphkv import nflx_graphkv_capacity_model
from .kafka import nflx_kafka_capacity_model
from .key_value import nflx_key_value_capacity_model
from .read_only_kv import nflx_read_only_kv_capacity_model
from .postgres import nflx_postgres_capacity_model
from .rds import nflx_rds_capacity_model
from .stateless_java import nflx_java_app_capacity_model
from .time_series import nflx_time_series_capacity_model
from .wal import nflx_wal_capacity_model
from .zookeeper import nflx_zookeeper_capacity_model


def models() -> Dict[str, Any]:
    return {
        "org.netflix.cassandra": nflx_cassandra_capacity_model,
        "org.netflix.stateless-java": nflx_java_app_capacity_model,
        "org.netflix.key-value": nflx_key_value_capacity_model,
        "org.netflix.time-series": nflx_time_series_capacity_model,
        "org.netflix.counter": nflx_counter_capacity_model,
        "org.netflix.zookeeper": nflx_zookeeper_capacity_model,
        "org.netflix.evcache": nflx_evcache_capacity_model,
        "org.netflix.elasticsearch": nflx_elasticsearch_capacity_model,
        "org.netflix.elasticsearch.node": nflx_elasticsearch_data_capacity_model,
        "org.netflix.elasticsearch.master": nflx_elasticsearch_master_capacity_model,
        "org.netflix.elasticsearch.search": nflx_elasticsearch_search_capacity_model,
        "org.netflix.entity": nflx_entity_capacity_model,
        "org.netflix.control": nflx_control_capacity_model,
        "org.netflix.cockroachdb": nflx_cockroachdb_capacity_model,
        "org.netflix.aurora": nflx_aurora_capacity_model,
        "org.netflix.postgres": nflx_postgres_capacity_model,
        "org.netflix.rds": nflx_rds_capacity_model,
        "org.netflix.kafka": nflx_kafka_capacity_model,
        "org.netflix.dynamodb": nflx_ddb_capacity_model,
        "org.netflix.wal": nflx_wal_capacity_model,
        "org.netflix.graphkv": nflx_graphkv_capacity_model,
        "org.netflix.read-only-kv": nflx_read_only_kv_capacity_model,
    }



# Contains definitions for AWS instance families.
# These are _relative_ IPC values with Skylake as a baseline.
# They are not targeted to a specific workload, but instead an average
# typical value across a broad range of tests such as those used
# in common online benchmarks.
#
from typing import Any
from typing import Dict

# Intel
HASWELL_IPC = 0.85
SKYLAKE_IPC = 1.0
ICELAKE_IPC = SKYLAKE_IPC * 1.15
SAPPHIRE_RAPIDS_IPC = ICELAKE_IPC * 1.12
EMERALD_RAPIDS_IPC = SAPPHIRE_RAPIDS_IPC * 1.03
GRANITE_RAPIDS_IPC = SAPPHIRE_RAPIDS_IPC * 1.10

# AMD
ROME_IPC = 1.03
MILAN_IPC = SKYLAKE_IPC * 1.15
GENOA_IPC = MILAN_IPC * 1.13

INSTANCE_TYPES: Dict[str, Dict[str, Any]] = {
    "c5": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.9,
        "cpu_turbo_all_ghz": 3.6,
    },
    "c5a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": ROME_IPC,
    },  # no spinnaker
    "c5d": {
        "xl_iops": None,
        "io_latency_curve": "5th-gen-ssd",
        "cpu_ipc_scale": SKYLAKE_IPC,
    },  # no spinnaker
    "c5n": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.4,
    },  # no spinnaker
    "c6a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": MILAN_IPC,
    },  # no spinnaker
    "c6i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "c6id": {
        "xl_iops": None,
        "io_latency_curve": "6th-gen-ssd",
        "cpu_ipc_scale": ICELAKE_IPC,
    },  # no spinnaker
    "c7a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GENOA_IPC,
        "cpu_turbo_single_ghz": 3.7,
        "cpu_turbo_all_ghz": 3.7,
    },
    "c7i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SAPPHIRE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.8,
        "cpu_turbo_all_ghz": 3.2,
    },
    "c8i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GRANITE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.9,
        "cpu_turbo_all_ghz": 3.9,
    },
    # "g4ad": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None},
    # "g4dn": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None},
    # "g5": {'xl_iops': None, 'io_latency_curve': '5th-gen-ssd', 'cpu_ipc_scale': None},
    # "g6": {'xl_iops': None, 'io_latency_curve': '6th-gen-ssd', 'cpu_ipc_scale': None},
    # "g6e": {
    #     'xl_iops': None,
    #     'io_latency_curve': '6th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    # "hpc7g": {
    #     'xl_iops': None,
    #     'io_latency_curve': '6th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    # "i3": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None},
    "i7i": {
        "xl_iops": "150,000/82,500",
        "io_latency_curve": "7th-gen-ephemeral",
        "cpu_ipc_scale": EMERALD_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 4.0,
        "cpu_turbo_all_ghz": 3.2,
    },
    "i3en": {
        "xl_iops": None,
        "io_latency_curve": "ssd",
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.1,
        "cpu_turbo_all_ghz": 3.1,
    },
    "i4i": {
        "xl_iops": None,
        "io_latency_curve": "ssd",
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "m4": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": HASWELL_IPC,
        "cpu_turbo_single_ghz": 3.0,
        "cpu_turbo_all_ghz": 2.6,
    },  # all-core is a guess
    "m5": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.1,
    },
    # exclude m5d and m5dn because they are in the manual list
    # "m5d": {
    #     'xl_iops': None,
    #     'io_latency_curve': '5th-gen-ssd',
    #     'cpu_ipc_scale': SKYLAKE_IPC,
    #     'cpu_turbo_single_ghz': 3.5,
    #     'cpu_turbo_all_ghz': 3.1
    # },
    # "m5dn": {
    #     'xl_iops': None,
    #     'io_latency_curve': '5th-gen-ssd',
    #     'cpu_ipc_scale': SKYLAKE_IPC,
    #     'cpu_turbo_single_ghz': 3.5,
    #     'cpu_turbo_all_ghz': 3.1
    # },
    "m5n": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.1,
    },
    "m6a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": MILAN_IPC,
        "cpu_turbo_single_ghz": 3.6,
        "cpu_turbo_all_ghz": 3.6,
    },
    # "m6gd": {
    #     'xl_iops': None,
    #     'io_latency_curve': '6th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    "m6i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "m6id": {
        "xl_iops": None,
        "io_latency_curve": "6th-gen-ssd",
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "m6idn": {
        "xl_iops": None,
        "io_latency_curve": "6th-gen-ssd",
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "m7a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GENOA_IPC,
        "cpu_turbo_single_ghz": 3.7,
        "cpu_turbo_all_ghz": 3.7,
    },  # is this turbo speed correct?
    "m7i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SAPPHIRE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.8,
        "cpu_turbo_all_ghz": 3.2,
    },
    "m8i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GRANITE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.9,
        "cpu_turbo_all_ghz": 3.9,
    },
    # "mac2-m2pro": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None},
    # "p4d": {
    #     'xl_iops': None,
    #     'io_latency_curve': '5th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    # "p4de": {
    #     'xl_iops': None,
    #     'io_latency_curve': '5th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    # "p5": {
    #     'xl_iops': None,
    #     'io_latency_curve':'6th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    # "p5en": {
    #     'xl_iops': None,
    #     'io_latency_curve': '6th-gen-ssd',
    #     'cpu_ipc_scale': None
    # },
    "r4": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": HASWELL_IPC,
        "cpu_turbo_single_ghz": 3.0,
        "cpu_turbo_all_ghz": 2.6,
    },  # all-core is a guess
    "r5": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.1,
    },
    # exclude r5d and r5dn because they are in the manual list
    # "r5d": {
    #     "xl_iops": None,
    #     "io_latency_curve": "5th-gen-ssd",
    #     "cpu_ipc_scale": SKYLAKE_IPC,
    #     "cpu_turbo_single_ghz": 3.5,
    #     "cpu_turbo_all_ghz": 3.1,
    # },
    # "r5dn": {
    #     "xl_iops": None,
    #     "io_latency_curve": "5th-gen-ssd",
    #     "cpu_ipc_scale": SKYLAKE_IPC,
    #     "cpu_turbo_single_ghz": 3.5,
    #     "cpu_turbo_all_ghz": 3.1,
    # },
    "r5n": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SKYLAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.1,
    },
    "r6a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": MILAN_IPC,
        "cpu_turbo_single_ghz": 3.6,
        "cpu_turbo_all_ghz": 3.6,
    },
    "r6i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "r6id": {
        "xl_iops": None,
        "io_latency_curve": "6th-gen-ssd",
        "cpu_ipc_scale": ICELAKE_IPC,
        "cpu_turbo_single_ghz": 3.5,
        "cpu_turbo_all_ghz": 3.5,
    },
    "r7a": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GENOA_IPC,
        "cpu_turbo_single_ghz": 3.7,
        "cpu_turbo_all_ghz": 3.7,
    },
    "r7i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": SAPPHIRE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.8,
        "cpu_turbo_all_ghz": 3.2,
    },
    "r8i": {
        "xl_iops": None,
        "io_latency_curve": None,
        "cpu_ipc_scale": GRANITE_RAPIDS_IPC,
        "cpu_turbo_single_ghz": 3.9,
        "cpu_turbo_all_ghz": 3.9,
    },
    # "t3": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None},
    # "z1d": {'xl_iops': None, 'io_latency_curve': 'ssd', 'cpu_ipc_scale': None}
}



import logging
import math
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionClusterCapacity
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import gp2_gib_for_io
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores

logger = logging.getLogger(__name__)


def _estimate_rds_requirement(
    instance: Instance, desires: CapacityDesires, db_type: str
) -> CapacityRequirement:
    """Estimate the capacity required for one region given a regional desire. Unlike
    Cassandra RDS instances are deployed per region, not zone. The input desires
    should be the **regional** desire, and this function will return the regional
    capacity requirement
    """

    if db_type == "postgres":
        needed_cores = sqrt_staffed_cores(desires) * 1.2  # 20% head room for VACUUM
    else:
        needed_cores = sqrt_staffed_cores(desires) * 1.1  # Just a guess!

    needed_cores = normalize_cores(
        core_count=needed_cores,
        target_shape=instance,
        reference_shape=desires.reference_shape,
    )

    # 20% head room For replication, backups etc.
    needed_network_mbps = simple_network_mbps(desires) * 1.2
    needed_disk = round(desires.data_shape.estimated_state_size_gib.mid, 2)

    if (
        desires.data_shape.estimated_working_set_percent
        and desires.data_shape.estimated_working_set_percent.mid
    ):
        working_set_percent = desires.data_shape.estimated_working_set_percent.mid
    else:
        working_set_percent = 0.10

    needed_memory = (
        desires.data_shape.estimated_state_size_gib.mid * working_set_percent
    )

    return CapacityRequirement(
        requirement_type="rds-regional",
        reference_shape=desires.reference_shape,
        cpu_cores=certain_int(needed_cores),
        mem_gib=certain_float(needed_memory),
        disk_gib=certain_float(needed_disk),
        network_mbps=certain_float(needed_network_mbps),
    )


# MySQL default block size is 16KiB, PostGreSQL is 8KiB Number of reads for B-Tree
# are given by log of total pages to the base of B-Tree fan out factor
def _rds_required_disk_ios(
    disk_size_gib: int, db_type: str, btree_fan_out: int = 100
) -> float:
    disk_size_kb = disk_size_gib * 1024 * 1024
    if db_type == "postgres":
        default_block_size = 8  # KiB
    else:
        default_block_size = 16  # MySQL default block size in KiB

    pages = max(1, disk_size_kb // default_block_size)
    return math.log(pages, btree_fan_out)


def _compute_rds_region(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    needed_cores: int,
    needed_disk_gib: int,
    needed_memory_gib: int,
    needed_network_mbps: float,
    required_disk_ios: Callable[[int], float],
    required_disk_space: Callable[[int], int],
    reference_shape: Instance,
) -> Optional[RegionClusterCapacity]:
    """Computes a regional cluster of a RDS service

    Basically just verifies that a single instance of a passed in instance type can
    support required cpu, memory and network since we can't scale RDS horizontally by
    adding more instances like Cassandra. Count of instance is always 1
    """

    needed_cores = normalize_cores(
        core_count=needed_cores, target_shape=instance, reference_shape=reference_shape
    )

    # We can't scale RDS horizontally by adding more nodes like we can for Cassandra
    # so single instance must meet the whole cpu, disk, memory and network bandwidth
    # requirement
    if (
        instance.cpu < needed_cores
        or instance.ram_gib < needed_memory_gib
        or instance.net_mbps < needed_network_mbps
    ):
        return None

    # calculate storage cost
    attached_drives = []
    space_gib = max(1, required_disk_space(needed_disk_gib))
    io_gib = gp2_gib_for_io(required_disk_ios(needed_disk_gib))
    rds_gib = max(io_gib, space_gib)
    attached_drive = drive.model_copy()
    attached_drive.size_gib = rds_gib
    attached_drives.append(attached_drive)
    total_annual_cost = instance.annual_cost + attached_drive.annual_cost

    logger.debug(
        "For (cpu, memory_gib, disk_gib) = (%s, %s, %s) need ( %s, %s, %s)",
        needed_cores,
        needed_memory_gib,
        needed_disk_gib,
        instance.name,
        attached_drives,
        total_annual_cost,
    )

    return RegionClusterCapacity(
        cluster_type="rds-cluster",
        count=1,
        instance=instance,
        attached_drives=attached_drives,
        annual_cost=total_annual_cost,
    )


def _estimate_rds_regional(
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    extra_model_arguments: Dict[str, Any],
) -> Optional[CapacityPlan]:
    instance_family = instance.family
    if instance_family not in ("m5", "r5"):
        return None

    if drive.name != "gp2":
        return None

    # RDS cannot make tier 0 service guarantees in terms partitions or availability
    if desires.service_tier == 0:
        return None

    db_type = extra_model_arguments.get("rds.engine", "mysql")
    requirement = _estimate_rds_requirement(instance, desires, db_type)
    rps = desires.query_pattern.estimated_read_per_second.mid

    cluster: Optional[RegionClusterCapacity] = _compute_rds_region(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        required_disk_ios=lambda size_gib: _rds_required_disk_ios(size_gib, db_type)
        * math.ceil(0.1 * rps),
        required_disk_space=lambda x: math.ceil(x * 1.2),  # Unscientific random guess!
        reference_shape=desires.reference_shape,
    )

    if not cluster:
        return None

    if desires.service_tier < 3:
        replicas = 2
    else:
        replicas = 1

    costs = {"rds-cluster.regional-clusters": cluster.annual_cost * replicas}
    clusters = Clusters(
        annual_costs=costs,
        zonal=[],
        regional=[cluster],
    )

    return CapacityPlan(
        requirements=Requirements(
            zonal=[],
            regional=[requirement] * replicas,
            regrets=("spend", "disk", "mem"),
        ),
        candidate_clusters=clusters,
    )


class NflxRDSArguments(BaseModel):
    rds_engine: str = Field(
        alias="rds.engine",
        default="mysql",
        description="RDS Database type",
    )


class NflxRDSCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        return _estimate_rds_regional(
            instance=instance,
            drive=drive,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        )

    @staticmethod
    def description() -> str:
        return "Netflix RDS Cluster Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxRDSArguments.model_json_schema()

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.latency,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.serializable_stale,
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.never,
                    ),
                ),
                # can't really make latency/throughput trade-offs with RDS
                estimated_mean_read_size_bytes=Interval(
                    low=128, mid=1024, high=65536, confidence=0.90
                ),
                estimated_mean_write_size_bytes=Interval(
                    low=64, mid=512, high=2048, confidence=0.90
                ),
                # probably closer to CRDB than Cassandra. Query by PK in MySQL takes
                # total of ~300 ms end to end
                estimated_mean_read_latency_ms=Interval(
                    low=5, mid=30, high=150, confidence=0.90
                ),
                estimated_mean_write_latency_ms=Interval(
                    low=5, mid=30, high=150, confidence=0.90
                ),
                # Single row fetch by PK in MySQL takes total of ~300 ms end to end
                read_latency_slo_ms=FixedInterval(
                    minimum_value=10,
                    maximum_value=2000,
                    low=300,
                    mid=800,
                    high=2000,
                    confidence=0.90,
                ),
                write_latency_slo_ms=FixedInterval(
                    minimum_value=10,
                    maximum_value=2000,
                    low=300,
                    mid=800,
                    high=2000,
                    confidence=0.90,
                ),
            ),
            data_shape=DataShape(
                estimated_working_set_percent=Interval(
                    low=0.05, mid=0.50, high=0.70, confidence=0.8
                )
            ),
        )


nflx_rds_capacity_model = NflxRDSCapacityModel()



from datetime import timedelta
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from pydantic import Field

from .stateless_java import nflx_java_app_capacity_model
from .stateless_java import NflxJavaAppArguments
from service_capacity_modeling.enum_utils import StrEnum
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.models import CapacityModel


class NflxCounterCardinality(StrEnum):
    low = "low"
    medium = "medium"
    high = "high"


class NflxCounterMode(StrEnum):
    best_effort = "best-effort"
    eventual = "eventual"
    exact = "exact"


class NflxCounterArguments(NflxJavaAppArguments):
    counter_global: bool = Field(
        alias="counter.global",
        description="Indicate if this use case requires global counts which "
        "is more expensive than regional",
    )
    counter_cardinality: NflxCounterCardinality = Field(
        alias="counter.cardinality",
        description="Low means < 10,000, medium (10,000—1,000,000), high means "
        "> 1,000,000.",
    )
    counter_mode: NflxCounterMode = Field(
        alias="counter.mode",
        description="What mode of counting",
    )


class NflxCounterCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # Counter wants 20GiB root volumes
        extra_model_arguments.setdefault("root_disk_gib", 20)

        counter_app = nflx_java_app_capacity_model.capacity_plan(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            extra_model_arguments=extra_model_arguments,
        )
        if counter_app is None:
            return None

        for cluster in counter_app.candidate_clusters.regional:
            cluster.cluster_type = "dgwcounter"
        return counter_app

    @staticmethod
    def description() -> str:
        return "Netflix Streaming Counter Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxCounterArguments.model_json_schema()

    @staticmethod
    def compose_with(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> Tuple[Tuple[str, Callable[[CapacityDesires], CapacityDesires]], ...]:
        stores = []

        if extra_model_arguments["counter.mode"] == NflxCounterMode.best_effort:
            stores.append(("org.netflix.evcache", lambda x: x))
        else:
            # Shared evcache cluster is used for eventual and exact counters
            def _modify_cassandra_desires(
                user_desires: CapacityDesires,
            ) -> CapacityDesires:
                modified = user_desires.model_copy(deep=True)
                counter_cardinality = extra_model_arguments["counter.cardinality"]

                counter_deltas_per_second = (
                    user_desires.query_pattern.estimated_write_per_second
                )

                # low cardinality : rollups happen once every 60 seconds
                # medium cardinality : rollups happen once every 30 seconds
                # high cardinality : rollups happen once every 10 seconds
                # TODO: Account for read amplification from time slice configs
                #       for better model accuracy
                if counter_cardinality == NflxCounterCardinality.low:
                    rollups_per_second = counter_deltas_per_second.scale(0.0167)
                elif counter_cardinality == NflxCounterCardinality.medium:
                    rollups_per_second = counter_deltas_per_second.scale(0.0333)
                else:
                    rollups_per_second = counter_deltas_per_second.scale(0.1)

                modified.query_pattern.estimated_read_per_second = rollups_per_second

                # storage size fix
                delta_event_size = 256  # bytes
                rolled_up_count_size = 128  # bytes
                GiB = 1024 * 1024 * 1024

                # Events can be discarded as soon as rollup is complete
                # We default to a 1 day slice with 2 day retention
                retention = timedelta(days=2).total_seconds()

                cardinality = {
                    "low": 10_000,
                    "medium": 100_000,
                    "high": 1_000_000,
                }[extra_model_arguments["counter.cardinality"]]

                event_storage_size = counter_deltas_per_second.scale(
                    delta_event_size * retention / GiB
                )
                rollup_storage_size = rolled_up_count_size * cardinality / GiB
                total_store_size = event_storage_size.offset(rollup_storage_size)
                modified.data_shape.estimated_state_size_gib = total_store_size

                return modified

            stores.append(("org.netflix.cassandra", _modify_cassandra_desires))
        return tuple(stores)

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=1,
                        mid=3,
                        high=6,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=50, high=200, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    # counter scan queries can be more expensive
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=4, high=6, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # Assume they're doing GetItems scans -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=1,
                        mid=8,
                        high=90,
                        confidence=0.98,
                    ),
                    # Assume they're doing PutRecords (BATCH)
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=20,
                        low=2,
                        mid=4,
                        high=10,
                        confidence=0.98,
                    ),
                ),
                # Most throughput counter clusters are large
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                    reserved_instance_app_mem_gib=8,
                ),
            )


nflx_counter_capacity_model = NflxCounterCapacityModel()



import json
from decimal import Decimal

from service_capacity_modeling.interface import Clusters


def test_total_annual_cost():
    """make sure total_annual_cost is calculated and dumped correctly"""
    cluster = Clusters(
        annual_costs={"right-zonal": Decimal(1234), "right-regional": Decimal(234)}
    )
    expected_total = float(cluster.total_annual_cost)

    assert expected_total == cluster.model_dump().get("total_annual_cost")
    assert expected_total == cluster.model_dump().get("total_annual_cost")
    assert expected_total == json.loads(cluster.model_dump_json()).get(
        "total_annual_cost"
    )
    assert expected_total == json.loads(cluster.model_dump_json()).get(
        "total_annual_cost"
    )



import math
from typing import Any
from typing import Dict
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import Service
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.models import CapacityModel

_TIER_TARGET_UTILIZATION_MAPPING = {
    0: 0.20,
    1: 0.40,
    2: 0.60,
    3: 0.80,
}


class NflxDynamoDBArguments(BaseModel):
    transactional_write_percent: float = Field(
        default=0,
        description="Coordinated writes, if multiple items in single write "
        "then all-or-nothing guarantee for writes. "
        "Default is zero, assuming most writes are single item puts",
    )
    transactional_read_percent: float = Field(
        default=0,
        description="Coordinated reads, if multiple items in "
        "single read then all-or-nothing guarantee for read. "
        "Default is zero",
    )
    target_max_annual_cost: float = Field(
        default=0,
        description="Target to determine max capacity units for auto-scale. "
        "Model does best effort to predict the capacity units but "
        "does not guarantee that actual costs will be less than "
        "what is specified here. Base costs based on provided requirements "
        "itself can exceed the target, if base costs are lower then "
        "actual costs can be 2x-3x of target",
    )


class _ReadPlan(BaseModel):
    read_capacity_units: int = Field(
        default=0,
        description="read capacity units needed to support the requested reads",
    )
    total_annual_read_cost: float = Field(
        default=0,
        description="annual cost to consume the provisioned read capacity units",
    )


class _WritePlan(BaseModel):
    write_capacity_units: int = Field(
        default=0,
        description="write capacity units needed to support the requested writes",
    )
    replicated_write_capacity_units: int = Field(
        default=0,
        description="write capacity units needed to "
        "support the requested writes for global tables",
    )
    total_annual_write_cost: float = Field(
        default=0,
        description="annual cost to consume the provisioned write capacity units",
    )


class _DataStoragePlan(BaseModel):
    total_data_storage_gib: float = Field(
        default=0,
        description="total amount of data stored in gib",
    )
    total_annual_data_storage_cost: float = Field(
        default=0,
        description="annual cost to store the requested data",
    )


class _DataBackupPlan(BaseModel):
    total_backup_data_storage_gib: float = Field(
        default=0,
        description="total amount of data stored for backup in gib",
    )
    total_annual_backup_cost: float = Field(
        default=0,
        description="annual cost to backup the stored data",
    )


class _DataTransferPlan(BaseModel):
    total_data_transfer_gib: float = Field(
        default=0,
        description="amount of data transferred, "
        "included the global table cross region replication",
    )
    total_annual_data_transfer_cost: float = Field(
        default=0,
        description="annual cost for data transfer",
    )


def _get_read_consistency_percentages(
    desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
) -> Dict[str, float]:
    # either all the required attributes are supplied in extra_model_arguments
    # or is derived based on the access consistency
    eventual_read_percent = extra_model_arguments.get("eventual_read_percent", 0.0)
    assert eventual_read_percent >= 0.0
    transactional_read_percent = extra_model_arguments.get(
        "transactional_read_percent", 0.0
    )
    assert transactional_read_percent >= 0.0
    strong_read_percent = extra_model_arguments.get("strong_read_percent", 0.0)
    assert strong_read_percent >= 0.0
    total_percent = (
        eventual_read_percent + transactional_read_percent + strong_read_percent
    )
    if total_percent == 0:
        access_consistency = desires.query_pattern.access_consistency.same_region
        target_consistency = access_consistency.target_consistency
        if target_consistency == AccessConsistency.serializable:
            transactional_read_percent = 1.0
            eventual_read_percent = 0.0
            strong_read_percent = 0.0
        elif target_consistency in (
            AccessConsistency.read_your_writes,
            AccessConsistency.linearizable,
        ):
            transactional_read_percent = 0.0
            eventual_read_percent = 0.0
            strong_read_percent = 1.0
        else:
            transactional_read_percent = 0.0
            eventual_read_percent = 1.0
            strong_read_percent = 0.0
    total_percent = (
        eventual_read_percent + transactional_read_percent + strong_read_percent
    )
    assert total_percent == 1, (
        "eventual_read_percent, transactional_read_percent, strong_read_percent"
        " should sum to 1"
    )
    return {
        "transactional_read_percent": transactional_read_percent,
        "eventual_read_percent": eventual_read_percent,
        "strong_read_percent": strong_read_percent,
    }


def _get_write_consistency_percentages(
    desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
) -> Dict[str, float]:
    # either all the required attributes are supplied in extra_model_arguments
    # or is derived based on the access consistency
    transactional_write_percent: float = extra_model_arguments.get(
        "transactional_write_percent", 0
    )
    assert transactional_write_percent >= 0.0
    non_transactional_write_percent: float = extra_model_arguments.get(
        "non_transactional_write_percent", 0
    )
    assert non_transactional_write_percent >= 0.0
    total_percent = transactional_write_percent + non_transactional_write_percent
    if total_percent == 0:
        access_consistency = desires.query_pattern.access_consistency.same_region
        target_consistency = access_consistency.target_consistency
        if target_consistency == AccessConsistency.serializable:
            transactional_write_percent = 1.0
            non_transactional_write_percent = 0.0
        else:
            transactional_write_percent = 0.0
            non_transactional_write_percent = 1.0
    total_percent = transactional_write_percent + non_transactional_write_percent
    assert total_percent == 1, (
        "transactional_write_percent, non_transactional_write_percent should sum to 1"
    )
    return {
        "transactional_write_percent": transactional_write_percent,
        "non_transactional_write_percent": non_transactional_write_percent,
    }


def _mean_write_item_size_bytes(desires: CapacityDesires) -> float:
    mean_item_size = desires.query_pattern.estimated_mean_write_size_bytes.mid
    return mean_item_size


def _mean_read_item_size_bytes(desires: CapacityDesires) -> float:
    mean_item_size = desires.query_pattern.estimated_mean_read_size_bytes.mid
    return mean_item_size


def _get_dynamo_standard(context: RegionContext) -> Service:
    number_of_regions = context.num_regions
    dynamo_service = (
        context.services.get("dynamo.standard.global", None)
        if number_of_regions > 1
        else context.services.get("dynamo.standard", None)
    )
    if not dynamo_service:
        raise ValueError("DynamoDB Service is not available in context")
    return dynamo_service


def _get_dynamo_backup(context: RegionContext) -> Service:
    dynamo_backup = context.services.get("dynamo.backup.continuous", None)
    if not dynamo_backup:
        raise ValueError("DynamoDB Backup is not available in context")
    return dynamo_backup


def _get_dynamo_transfer(
    context: RegionContext,
) -> Service:
    transfer_costs = context.services.get("dynamo.transfer", None)
    if not transfer_costs:
        raise ValueError("DynamoDB Transfer cost is not available in context")
    return transfer_costs


def _plan_writes(
    context: RegionContext,
    desires: CapacityDesires,
    extra_model_arguments: Dict[str, Any],
) -> _WritePlan:
    mean_item_size = _mean_write_item_size_bytes(desires)

    # For items up to 1 KB in size,
    # one WCU can perform one standard write request per second
    rounded_wcus_per_item = math.ceil(max(1.0, mean_item_size / 1024))

    write_percentages = _get_write_consistency_percentages(
        desires, extra_model_arguments
    )

    transactional_write_percent = write_percentages["transactional_write_percent"]
    baseline_wcus_non_transactional = (
        desires.query_pattern.estimated_write_per_second.mid
        * (1 - transactional_write_percent)
        * rounded_wcus_per_item
    )

    # Transactional write requests require two WCUs
    baseline_wcus_transactional = (
        desires.query_pattern.estimated_write_per_second.mid
        * transactional_write_percent
        * 2
        * rounded_wcus_per_item
    )
    # we do not break down reserved vs non-reserved
    total_baseline_wcus = math.ceil(
        baseline_wcus_non_transactional + baseline_wcus_transactional
    )

    # 8760 hours in a year (365 * 24)
    total_baseline_wcus_hours = 8760 * total_baseline_wcus
    number_of_regions = context.num_regions

    # pick the right pricing based on context
    dynamo_service_standard = _get_dynamo_standard(context)

    # for global replication each region will additionally consume rWCUs
    total_annual_cost_wcu = round(
        (
            number_of_regions
            * total_baseline_wcus_hours
            * dynamo_service_standard.annual_cost_per_write_io
        ),
        2,
    )
    if number_of_regions > 1:
        return _WritePlan(
            replicated_write_capacity_units=total_baseline_wcus_hours,
            write_capacity_units=0,
            total_annual_write_cost=total_annual_cost_wcu,
        )
    return _WritePlan(
        write_capacity_units=total_baseline_wcus_hours,
        replicated_write_capacity_units=0,
        total_annual_write_cost=total_annual_cost_wcu,
    )


def _plan_reads(
    context: RegionContext,
    desires: CapacityDesires,
    extra_model_arguments: Dict[str, Any],
) -> _ReadPlan:
    read_percentages = _get_read_consistency_percentages(desires, extra_model_arguments)
    transactional_read_percent = read_percentages["transactional_read_percent"]
    eventual_read_percent = read_percentages["eventual_read_percent"]
    strong_read_percent = read_percentages["strong_read_percent"]
    mean_item_size = _mean_read_item_size_bytes(desires)

    # items up to 4 KB in size
    rounded_rcus_per_item = math.ceil(max(1.0, mean_item_size / (4 * 1024)))
    estimated_read_per_second = desires.query_pattern.estimated_read_per_second.mid

    # one RCU can perform two eventually consistent read
    baseline_rcus_eventually_consistent = (
        estimated_read_per_second * eventual_read_percent * 0.5 * rounded_rcus_per_item
    )

    # one RCU can perform one strongly consistent read
    baseline_rcus_strong_consistent = (
        estimated_read_per_second * strong_read_percent * 1 * rounded_rcus_per_item
    )

    # two RCUs can perform one transactional read
    baseline_rcus_transact = (
        estimated_read_per_second
        * transactional_read_percent
        * 2
        * rounded_rcus_per_item
    )

    total_baseline_rcus = math.ceil(
        baseline_rcus_eventually_consistent
        + baseline_rcus_strong_consistent
        + baseline_rcus_transact
    )

    # 8760 hours in a year (365 * 24)
    total_baseline_rcus_hours = 8760 * total_baseline_rcus
    dynamo_service_standard = _get_dynamo_standard(context)
    total_annual_cost_rcu_hour = round(
        (total_baseline_rcus_hours * dynamo_service_standard.annual_cost_per_read_io), 2
    )
    return _ReadPlan(
        read_capacity_units=total_baseline_rcus_hours,
        total_annual_read_cost=total_annual_cost_rcu_hour,
    )


def _plan_storage(
    context: RegionContext,
    desires: CapacityDesires,
) -> _DataStoragePlan:
    dynamo_service_standard = _get_dynamo_standard(context)
    annual_storage_cost = round(
        (
            dynamo_service_standard.annual_cost_gib(
                math.ceil(desires.data_shape.estimated_state_size_gib.mid)
            )
        ),
        2,
    )
    return _DataStoragePlan(
        total_data_storage_gib=round(
            desires.data_shape.estimated_state_size_gib.mid, 2
        ),
        total_annual_data_storage_cost=annual_storage_cost,
    )


def _plan_data_transfer(
    context: RegionContext,
    desires: CapacityDesires,
) -> _DataTransferPlan:
    number_of_regions = context.num_regions
    if not number_of_regions > 1:
        return _DataTransferPlan(
            total_data_transfer_gib=0, total_annual_data_transfer_cost=0
        )
    mean_item_size_bytes = _mean_write_item_size_bytes(desires)
    writes_per_second = desires.query_pattern.estimated_write_per_second.mid
    # 31,536,000 seconds in a year (365 * 24 * 60 * 60)
    # 1024 * 1024 * 1024 = 1Gib
    annual_data_written_gib = round(
        ((31536000 / (1024 * 1024 * 1024)) * writes_per_second * mean_item_size_bytes),
        2,
    )
    transfer_costs = _get_dynamo_transfer(context)
    annual_transfer_cost_to_another_region = transfer_costs.annual_cost_gib(
        annual_data_written_gib
    )
    annual_transfer_cost = round(
        annual_transfer_cost_to_another_region * (number_of_regions - 1), 2
    )
    return _DataTransferPlan(
        total_data_transfer_gib=annual_data_written_gib * (number_of_regions - 1),
        total_annual_data_transfer_cost=annual_transfer_cost,
    )


def _plan_backup(
    context: RegionContext,
    desires: CapacityDesires,
) -> _DataBackupPlan:
    number_of_regions = context.num_regions
    dynamo_backup_continuous = _get_dynamo_backup(context)
    annual_pitr_cost = dynamo_backup_continuous.annual_cost_gib(
        math.ceil(desires.data_shape.estimated_state_size_gib.mid)
    )
    # normalizing the cost as pitr is not charged per region
    annual_pitr_cost_normalized = round((annual_pitr_cost / number_of_regions), 2)
    return _DataBackupPlan(
        total_backup_data_storage_gib=round(
            desires.data_shape.estimated_state_size_gib.mid, 2
        ),
        total_annual_backup_cost=annual_pitr_cost_normalized,
    )


class NflxDynamoDBCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # refer https://aws.amazon.com/dynamodb/pricing/provisioned/
        write_plan = _plan_writes(context, desires, extra_model_arguments)
        read_plan = _plan_reads(context, desires, extra_model_arguments)
        storage_plan = _plan_storage(context, desires)
        backup_plan = _plan_backup(context, desires)
        data_transfer_plan = _plan_data_transfer(context, desires)
        target_max_annual_cost: float = extra_model_arguments.get(
            "target_max_annual_cost", 0
        )
        target_util_percentage = 0.80
        if desires.service_tier in _TIER_TARGET_UTILIZATION_MAPPING:
            target_util_percentage = _TIER_TARGET_UTILIZATION_MAPPING[
                desires.service_tier
            ]

        requirement_context = {
            "read_capacity_units": read_plan.read_capacity_units,
            "write_capacity_units": write_plan.write_capacity_units,
            "data_transfer_gib": data_transfer_plan.total_data_transfer_gib,
            "target_utilization_percentage": target_util_percentage,
        }
        requirement_context["replicated_write_capacity_units"] = (
            write_plan.replicated_write_capacity_units
        )

        dynamo_costs = {
            "dynamo.regional-writes": write_plan.total_annual_write_cost,
            "dynamo.regional-reads": read_plan.total_annual_read_cost,
            "dynamo.regional-storage": storage_plan.total_annual_data_storage_cost,
        }

        dynamo_costs["dynamo.regional-transfer"] = (
            data_transfer_plan.total_annual_data_transfer_cost
        )

        dynamo_costs["dynamo.data-backup"] = backup_plan.total_annual_backup_cost

        total_annual_costs = round(sum(dynamo_costs.values()), 2)
        total_write_capacity_units = (
            write_plan.write_capacity_units + write_plan.replicated_write_capacity_units
        )
        max_read_capacity_units = max(1, read_plan.read_capacity_units)
        max_write_capacity_units = max(1, total_write_capacity_units)

        if target_max_annual_cost > 0:
            requirement_context["target_max_annual_cost"] = target_max_annual_cost
            annual_balance_target = target_max_annual_cost - total_annual_costs
            if annual_balance_target > 0:
                max_read_capacity_units += math.ceil(
                    (annual_balance_target / max(1.0, read_plan.total_annual_read_cost))
                    * read_plan.read_capacity_units,
                )
                max_write_capacity_units += math.ceil(
                    (
                        annual_balance_target
                        / max(1.0, write_plan.total_annual_write_cost)
                    )
                    * total_write_capacity_units,
                )

        requirement = CapacityRequirement(
            requirement_type="dynamo-regional",
            cpu_cores=certain_int(0),
            disk_gib=certain_float(storage_plan.total_data_storage_gib),
            context=requirement_context,
        )
        dynamo_services = [
            ServiceCapacity(
                service_type="dynamo.standard",
                annual_cost=(
                    write_plan.total_annual_write_cost
                    + read_plan.total_annual_read_cost
                    + storage_plan.total_annual_data_storage_cost
                    + data_transfer_plan.total_annual_data_transfer_cost
                ),
                service_params={
                    "read_capacity_units": {
                        "estimated": read_plan.read_capacity_units,
                        "auto_scale": {
                            "min": 1,
                            "max": max_read_capacity_units,
                            "target_utilization_percentage": target_util_percentage,
                        },
                    },
                    "write_capacity_units": {
                        "estimated": (
                            write_plan.write_capacity_units
                            + write_plan.replicated_write_capacity_units
                        ),
                        "auto_scale": {
                            "min": 1,
                            "max": max_write_capacity_units,
                            "target_utilization_percentage": target_util_percentage,
                        },
                    },
                },
            ),
            ServiceCapacity(
                service_type="dynamo.backup",
                annual_cost=backup_plan.total_annual_backup_cost,
            ),
        ]

        clusters = Clusters(
            annual_costs=dynamo_costs,
            zonal=[],
            regional=[],
            services=dynamo_services,
        )

        return CapacityPlan(
            requirements=Requirements(regional=[requirement]),
            candidate_clusters=clusters,
        )

    @staticmethod
    def description() -> str:
        return "Netflix Streaming DynamoDB Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxDynamoDBArguments.model_json_schema()

    @staticmethod
    def run_hardware_simulation() -> bool:
        return False

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        acceptable_consistency = {
            "same_region": {
                None,
                AccessConsistency.serializable,
                AccessConsistency.linearizable,
                AccessConsistency.best_effort,
                AccessConsistency.eventual,
                AccessConsistency.read_your_writes,
                AccessConsistency.never,
            },
            "cross_region": {
                None,
                # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/globaltables_HowItWorks.html#V2globaltables_HowItWorks.CommonTasks
                AccessConsistency.linearizable_stale,
                AccessConsistency.best_effort,
                AccessConsistency.eventual,
                AccessConsistency.never,
            },
        }

        for key, value in user_desires.query_pattern.access_consistency:
            if value.target_consistency not in acceptable_consistency[key]:
                raise ValueError(
                    f"DynamoDB can only provide {acceptable_consistency[key]} access."
                    f"User asked for {key}={value}"
                )

        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=256, high=1024, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=5, mid=10, high=15, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=5, mid=10, high=15, confidence=0.98
                    ),
                    # Assume point queries, "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=2,
                        maximum_value=20,
                        low=5,
                        mid=10,
                        high=15,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=2,
                        maximum_value=20,
                        low=5,
                        mid=10,
                        high=15,
                        confidence=0.98,
                    ),
                ),
                # Most latency sensitive clusters are in the
                # < 1TiB range
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=100, high=1000, confidence=0.98
                    ),
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=512, mid=2048, high=131072, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=10, mid=20, high=30, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=10, mid=20, high=30, confidence=0.98
                    ),
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=5,
                        maximum_value=100,
                        low=10,
                        mid=40,
                        high=90,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=5,
                        maximum_value=100,
                        low=10,
                        mid=40,
                        high=90,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                ),
            )


nflx_ddb_capacity_model = NflxDynamoDBCapacityModel()



import logging
import math
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.enum_utils import StrEnum
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import ClusterCapacity
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import CurrentZoneClusterCapacity
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import GIB_IN_BYTES
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import MEGABIT_IN_BYTES
from service_capacity_modeling.interface import MIB_IN_BYTES
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models import CostAwareModel
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import cluster_infra_cost
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import get_effective_disk_per_node_gib
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import zonal_requirements_from_current
from service_capacity_modeling.models.org.netflix.iso_date_math import iso_to_seconds

logger = logging.getLogger(__name__)


class ClusterType(StrEnum):
    strong = "strong"
    ha = "high-availability"


def _get_current_zonal_cluster(
    desires: CapacityDesires,
) -> Optional[CurrentZoneClusterCapacity]:
    return (
        None
        if desires.current_clusters is None
        else (
            desires.current_clusters.zonal[0]
            if len(desires.current_clusters.zonal)
            else None
        )
    )


def _is_same_instance_family(
    cluster: Optional[CurrentZoneClusterCapacity], target_family: str
) -> bool:
    """Check if cluster has a different instance family than the target."""
    if not cluster or not cluster.cluster_instance:
        return False
    return cluster.cluster_instance.family == target_family


def _estimate_kafka_requirement(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    desires: CapacityDesires,
    copies_per_region: int,
    hot_retention_seconds: float,
    zones_per_region: int = 3,
) -> Tuple[CapacityRequirement, Tuple[str, ...]]:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid
        / MIB_IN_BYTES
    )
    write_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid
        / MIB_IN_BYTES
    )

    # 1 concurrent reader, cpu time is per MiB read
    # 0.5 MiB / second
    rps = normalized_to_mib.query_pattern.estimated_read_per_second
    wps = normalized_to_mib.query_pattern.estimated_write_per_second
    normalized_to_mib.query_pattern.estimated_read_per_second = rps.scale(
        read_mib_per_second
    )
    normalized_to_mib.query_pattern.estimated_write_per_second = wps.scale(
        write_mib_per_second
    )
    # use the current cluster capacity if available
    current_zonal_capacity = _get_current_zonal_cluster(desires)

    bw_in = (
        (write_mib_per_second * MIB_IN_BYTES) * copies_per_region
    ) / MEGABIT_IN_BYTES
    bw_out = (
        (read_mib_per_second * MIB_IN_BYTES)
        + ((write_mib_per_second * MIB_IN_BYTES) * (copies_per_region - 1))
    ) / MEGABIT_IN_BYTES
    if (
        current_zonal_capacity
        and current_zonal_capacity.cluster_instance
        and desires.current_clusters is not None
    ):
        # zonal_requirements_from_current uses the midpoint utilization of the
        # current cluster. For Kafka, we want to use the high value instead
        # for cpu, disk, network, etc.
        normalize_midpoint_desires = desires.model_copy(deep=True)
        # This is checked in the if statement above. Assert here for mypy linter purpose
        assert normalize_midpoint_desires.current_clusters is not None
        curr_disk = desires.current_clusters.zonal[0].disk_utilization_gib
        curr_cpu = desires.current_clusters.zonal[0].cpu_utilization
        curr_network = desires.current_clusters.zonal[0].network_utilization_mbps
        normalize_midpoint_desires.current_clusters.zonal[
            0
        ].disk_utilization_gib = Interval(
            low=curr_disk.high, mid=curr_disk.high, high=curr_disk.high
        )
        normalize_midpoint_desires.current_clusters.zonal[0].cpu_utilization = Interval(
            low=curr_cpu.high, mid=curr_cpu.high, high=curr_cpu.high
        )
        normalize_midpoint_desires.current_clusters.zonal[
            0
        ].network_utilization_mbps = Interval(
            low=curr_network.high, mid=curr_network.high, high=curr_network.high
        )
        capacity_requirement = zonal_requirements_from_current(
            current_cluster=normalize_midpoint_desires.current_clusters,
            buffers=desires.buffers,
            instance=instance,
            reference_shape=current_zonal_capacity.cluster_instance,
        )
        needed_cores = int(capacity_requirement.cpu_cores.mid)
        needed_disk = int(capacity_requirement.disk_gib.mid)
        needed_network_mbps = int(capacity_requirement.network_mbps.mid)
        logger.debug(
            "kafka normalized needed cores: %s", capacity_requirement.cpu_cores
        )
    else:
        # We have no existing utilization to go from
        needed_cores = (
            normalize_cores(
                core_count=sqrt_staffed_cores(normalized_to_mib),
                target_shape=instance,
                reference_shape=desires.reference_shape,
            )
            // zones_per_region
        )

        # (Nick): Keep 40% of available bandwidth for node recovery
        # (Joey): For kafka BW = BW_write + BW_reads
        #   let X = input write BW
        #   BW_in = X * RF
        #   BW_out = X * (consumers) + X * (RF - 1)
        #   BW = (in + out) because duplex then 40% headroom.
        needed_network_mbps = int((max(bw_in, bw_out) * 1.40) // zones_per_region)

        # NOTE: data_shape is region, we need to convert it to zonal
        # If we don't have an existing cluster, the estimated state size should be
        # at most 40% of the total cluster's available disk.
        # i.e. needed_disk = state_size * 2.5
        needed_disk = math.ceil(
            (desires.data_shape.estimated_state_size_gib.mid // zones_per_region) * 2.5,
        )

    # Keep the last N seconds hot in cache
    needed_memory = (
        (write_mib_per_second * hot_retention_seconds) // 1024
    ) // zones_per_region

    logger.debug(
        "Need (instance, cpu, mem, disk) = (%s, %s, %s, %s)",
        instance.name,
        needed_cores,
        needed_memory,
        needed_disk,
    )

    return (
        CapacityRequirement(
            requirement_type="kafka-zonal",
            reference_shape=desires.reference_shape,
            cpu_cores=certain_int(needed_cores),
            mem_gib=certain_float(needed_memory),
            disk_gib=certain_float(needed_disk),
            network_mbps=certain_float(needed_network_mbps),
            context={
                "bw_in_mbps": bw_in,
                "bw_out_mbps": bw_out,
                "hot_retention_seconds": hot_retention_seconds,
                "replication_factor": copies_per_region,
            },
        ),
        ("spend", "disk", "mem"),
    )


def _upsert_params(cluster: Any, params: Dict[str, Any]) -> None:
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


def _kafka_read_io(
    rps: float, io_size_kib: float, size_gib: float, recovery_seconds: int
) -> float:
    # Get enough disk read IO capacity for some reads
    # In practice we have cache reducing this by 99% or more
    read_ios = rps * 0.05
    # Recover the node in 60 minutes, to do that we need
    # Estimate that we are using ~ 50% of the disk prior to a recovery
    size_kib = size_gib * 0.5 * (1024 * 1024)
    recovery_ios = max(1, size_kib / io_size_kib) / recovery_seconds
    # Leave 50% headroom for read IOs since generally we will hit cache
    return float((read_ios + int(round(recovery_ios))) * 1.5)


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-positional-arguments
def _estimate_kafka_cluster_zonal(  # noqa: C901
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    hot_retention_seconds: float,
    zones_per_region: int = 3,
    copies_per_region: int = 2,
    require_local_disks: bool = False,
    require_attached_disks: bool = False,
    required_zone_size: Optional[int] = None,
    max_regional_size: int = 150,
    max_local_data_per_node_gib: int = 2 * 1024,
    max_attached_data_per_node_gib: int = 2 * 1024,
    min_instance_cpu: int = 2,
    min_instance_memory_gib: int = 12,
    require_same_instance_family: bool = True,
) -> Optional[CapacityPlan]:
    # Kafka doesn't like to deploy on single CPU instances or with < 12 GiB of ram
    if instance.cpu < min_instance_cpu or instance.ram_gib < min_instance_memory_gib:
        return None

    # if we're not allowed to use attached disks, skip EBS only types
    if instance.drive is None and require_local_disks:
        return None

    # if we're not allowed to use local disks, skip ephems
    if instance.drive is not None and require_attached_disks:
        return None

    # Kafka only deploys on gp3 drives right now
    if instance.drive is None and drive.name != "gp3":
        return None

    # If there is a current cluster, check if we are restricted to same instance family
    current_zonal_cluster = _get_current_zonal_cluster(desires)
    if (
        current_zonal_cluster
        and require_same_instance_family
        and not _is_same_instance_family(current_zonal_cluster, instance.family)
    ):
        return None

    requirement, regrets = _estimate_kafka_requirement(
        instance=instance,
        desires=desires,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
        hot_retention_seconds=hot_retention_seconds,
    )

    # Account for sidecars and base system memory
    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    # Kafka clusters in prod (tier 0+1) need at least 2 nodes per zone
    min_count_for_tier = 1
    if desires.service_tier < 2:
        min_count_for_tier = 2

    # Kafka read io / second is zonal
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )
    write_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )

    # All Kafka IOs are sequential, so they can use the group size
    read_ios_per_second = max(1, (read_mib_per_second * 1024) // drive.seq_io_size_kib)
    write_ios_per_second = max(
        1, (write_mib_per_second * 1024) // drive.seq_io_size_kib
    )

    needed_disk_gib = int(requirement.disk_gib.mid)
    disk_buffer_ratio = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    ).ratio
    max_disk_per_node_gib = get_effective_disk_per_node_gib(
        instance,
        drive,
        disk_buffer_ratio,
        max_local_data_per_node_gib=max_local_data_per_node_gib,
        max_attached_data_per_node_gib=max_attached_data_per_node_gib,
    )
    min_count_for_data = math.ceil(needed_disk_gib / max_disk_per_node_gib)
    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # Leave 2x overhead for both reads and writes.
        # Readers are sequential, 100MiB/s read is 100Mib / block_size ios
        # Writers are sequential, 100MiB/s write is 100MiB / block_size ios
        required_disk_ios=lambda size, count: (
            _kafka_read_io(
                rps=read_ios_per_second / count,
                # Kafka does sequential IO
                io_size_kib=drive.seq_io_size_kib,
                size_gib=size,
                # Enough IO to recover a node in 60 minutes
                recovery_seconds=60 * 60,
            ),
            # Leave 100% IO headroom for writes
            copies_per_region * (write_ios_per_second / count) * 2,
        ),
        cluster_size=lambda x: x,
        min_count=max(min_count_for_tier, min_count_for_data, required_zone_size or 1),
        # Sidecars and Variable OS Memory
        # Kafka currently uses 8GiB fixed, might want to change to min(30, x // 2)
        reserve_memory=lambda instance_mem_gib: base_mem + 8,
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {"kafka.copies": copies_per_region}
    _upsert_params(cluster, params)

    # This is roughly the disk we would have tried to provision with the current
    # cluster's instance count (or required_zone_size)
    # If we *could* have satisified the required_zone_size without exceeding the
    # max data size per node constraints but the actual cluster count does not
    # match the required_zone_size, then we want to omit the plan
    # However if there was no way to satisfy the required_zone_size with the
    # max data size per node constraints, then we would rather return some result
    # with a higher instance count since this is the best we can do
    if (
        required_zone_size is not None
        and min_count_for_data <= required_zone_size
        and cluster.count != required_zone_size
    ):
        return None

    # Kafka clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * NxN network issues. Sometimes smaller clusters of bigger nodes
    #       are better for network propagation
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    cluster.cluster_type = NflxKafkaCapacityModel.cluster_type
    zonal_clusters = [cluster] * zones_per_region

    # Account for the clusters and replication costs
    kafka_costs = NflxKafkaCapacityModel.cluster_costs(
        service_type=NflxKafkaCapacityModel.service_name,
        zonal_clusters=zonal_clusters,
    )

    clusters = Clusters(
        annual_costs=kafka_costs,
        zonal=zonal_clusters,
        regional=[],
        services=[],
    )

    return CapacityPlan(
        requirements=Requirements(
            zonal=[requirement] * zones_per_region, regrets=regrets
        ),
        candidate_clusters=clusters,
    )


class NflxKafkaArguments(BaseModel):
    copies_per_region: int = Field(
        default=2,
        description=(
            "How many copies of the data will exist e.g. RF=3. If not supplied"
            " this will be determined from mode"
        ),
    )
    cluster_type: ClusterType = Field(
        default=ClusterType.ha,
        description="If the cluster is 'strong' consistency or 'high availability'",
    )
    retention: str = Field(
        default="PT8H", description="How long to retain data in this cluster."
    )
    hot_retention: str = Field(
        default="PT10M",
        description=(
            "How long to retain data in page cache for consumers to skip IO. "
            "Typically consumers lag under 10s, but ensure up to 10M can be handled"
        ),
    )
    require_local_disks: bool = Field(
        default=False,
        description="If local (ephemeral) drives are required",
    )
    require_attached_disks: bool = Field(
        default=False,
        description="If attached (ebs) drives are required",
    )
    required_zonal_size: Optional[int] = Field(
        default=None,
        description="Require zonal clusters to be this size (force vertical scaling)",
    )
    max_regional_size: int = Field(
        default=150,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=5120,
        description=(
            "The maximum amount of data we store per node. Used to limit "
            "recovery duration on failure."
        ),
    )
    min_instance_cpu: int = Field(
        default=2,
        description="The minimum number of instance CPU to allow",
    )
    min_instance_memory_gib: int = Field(
        default=12,
        description="The minimum amount of instance memory to allow",
    )


class NflxKafkaCapacityModel(CapacityModel, CostAwareModel):
    service_name = "kafka"
    cluster_type = "kafka"
    HA_DEFAULT_REPLICATION_FACTOR = 2
    SC_DEFAULT_REPLICATION_FACTOR = 3

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        cluster_type: ClusterType = ClusterType(
            extra_model_arguments.get("cluster_type", "high-availability")
        )
        replication_factor = NflxKafkaCapacityModel.HA_DEFAULT_REPLICATION_FACTOR
        if cluster_type == ClusterType.strong:
            replication_factor = NflxKafkaCapacityModel.SC_DEFAULT_REPLICATION_FACTOR
        copies_per_region: int = extra_model_arguments.get(
            "copies_per_region", replication_factor
        )
        if cluster_type == ClusterType.strong and copies_per_region < 3:
            raise ValueError("Strong consistency and RF<3 doesn't work")

        max_regional_size: int = extra_model_arguments.get("max_regional_size", 150)
        # Very large nodes are hard to cache warm
        max_local_data_per_node_gib: int = extra_model_arguments.get(
            "max_local_data_per_node_gib",
            extra_model_arguments.get("max_local_disk_gib", 1024 * 2),
        )
        max_attached_data_per_node_gib: int = extra_model_arguments.get(
            "max_attached_data_per_node_gib",
            extra_model_arguments.get("max_attached_disk_gib", 1024 * 2),
        )
        min_instance_cpu: int = extra_model_arguments.get("min_instance_cpu", 2)
        min_instance_memory_gib: int = extra_model_arguments.get(
            "min_instance_memory_gib", 12
        )
        hot_retention_seconds: float = iso_to_seconds(
            extra_model_arguments.get("hot_retention", "PT10M")
        )
        require_local_disks: bool = extra_model_arguments.get(
            "require_local_disks", False
        )
        require_attached_disks: bool = extra_model_arguments.get(
            "require_attached_disks", False
        )
        required_zone_size: Optional[int] = extra_model_arguments.get(
            "required_zone_size", None
        )
        # By default, for existing clusters, restrict to only using same instance family
        require_same_instance_family: bool = extra_model_arguments.get(
            "require_same_instance_family", True
        )

        return _estimate_kafka_cluster_zonal(
            instance=instance,
            drive=drive,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            require_local_disks=require_local_disks,
            require_attached_disks=require_attached_disks,
            required_zone_size=required_zone_size,
            max_regional_size=max_regional_size,
            max_local_data_per_node_gib=max_local_data_per_node_gib,
            max_attached_data_per_node_gib=max_attached_data_per_node_gib,
            min_instance_cpu=min_instance_cpu,
            min_instance_memory_gib=min_instance_memory_gib,
            hot_retention_seconds=hot_retention_seconds,
            require_same_instance_family=require_same_instance_family,
        )

    @staticmethod
    def cluster_costs(
        service_type: str,
        zonal_clusters: Sequence[ClusterCapacity] = (),
        regional_clusters: Sequence[ClusterCapacity] = (),
    ) -> Dict[str, float]:
        return cluster_infra_cost(
            service_type,
            zonal_clusters,
            regional_clusters,
            cluster_type=NflxKafkaCapacityModel.cluster_type,
        )

    @staticmethod
    def service_costs(
        service_type: str,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> List[ServiceCapacity]:
        _ = (service_type, context, desires, extra_model_arguments)
        return []

    @staticmethod
    def description() -> str:
        return "Netflix Streaming Kafka Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxKafkaArguments.model_json_schema()

    @staticmethod
    def allowed_cloud_drives() -> Tuple[Optional[str], ...]:
        # Kafka at Netflix only supports 3rd gen SSD storage
        return ("gp3",)

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        # Default to 10MiB/s and a single reader
        concurrent_readers = max(
            1, int(user_desires.query_pattern.estimated_read_per_second.mid)
        )
        query_pattern = user_desires.query_pattern.model_dump()
        if "estimated_mean_write_size_bytes" in query_pattern:
            write_bytes = Interval(**query_pattern["estimated_mean_write_size_bytes"])
        else:
            write_bytes = certain_int(10 * 1024 * 1024)

        read_bytes = write_bytes.scale(concurrent_readers)
        retention = extra_model_arguments.get("retention", "PT8H")
        retention_secs = iso_to_seconds(retention)

        # write throughput * retention * replication factor = usage
        replication_factor = NflxKafkaCapacityModel.HA_DEFAULT_REPLICATION_FACTOR
        if extra_model_arguments.get("cluster_type", None) == ClusterType.strong:
            replication_factor = NflxKafkaCapacityModel.SC_DEFAULT_REPLICATION_FACTOR
        state_gib = (
            write_bytes.mid * retention_secs * replication_factor
        ) / GIB_IN_BYTES

        # By supplying these buffers we can deconstruct observed utilization into
        # load versus buffer.
        compute_buffer_ratio = 2.5 if user_desires.service_tier in (0, 1) else 2.0
        buffers = Buffers(
            default=Buffer(ratio=1.5),
            desired={
                # Amount of compute buffer that we need to reserve in addition to
                # cpu_headroom_target that is reserved on a per instance basis
                "compute": Buffer(
                    ratio=compute_buffer_ratio, components=[BufferComponent.compute]
                ),
                # This makes sure we use only 40% of the available storage
                "storage": Buffer(ratio=2.5, components=[BufferComponent.storage]),
            },
        )

        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.throughput,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.read_your_writes
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.never
                    ),
                ),
                estimated_mean_write_size_bytes=write_bytes,
                estimated_mean_read_size_bytes=read_bytes,
                # How much on-cpu time to read or write 1 MiB
                # 42 MiB / second in 3 CPU seconds = 14 MiB / CPU second
                # This = 1000 / 14 = 71 millisecond / MiB
                # To read 1 MiB costs 71 millisecond of CPU time
                # Later tuned down based on training data
                estimated_mean_read_latency_ms=Interval(
                    low=20, mid=40, high=75, confidence=0.98
                ),
                # Wild-Ass-Guess that writes are slightly cheaper
                # Because no disks/context switching??
                estimated_mean_write_latency_ms=Interval(
                    low=20, mid=30, high=75, confidence=0.98
                ),
            ),
            data_shape=DataShape(
                estimated_state_size_gib=Interval(
                    low=state_gib * 0.5,
                    mid=state_gib,
                    high=state_gib * 2,
                    confidence=0.98,
                ),
                # Gandalf, system stuffs
                reserved_instance_app_mem_gib=1,
                # Connection overhead, kernel, etc ...
                reserved_instance_system_mem_gib=3,
            ),
            buffers=buffers,
        )


nflx_kafka_capacity_model = NflxKafkaCapacityModel()



"""Plan comparison utilities and types for capacity planning.

This module provides functionality to compare two capacity plans and determine
if they are roughly equivalent, with detailed explanations of any differences.

Example usage::

    from service_capacity_modeling.models.plan_comparison import (
        compare_plans,
        ignore_resource,
        lte,
        plus_or_minus,
        ResourceTolerances,
    )

    # Get recommendation from planner
    recommendations = planner.plan_certain(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=desires,
    )
    recommendation = recommendations[0]

    # Get current deployment as baseline using CapacityPlanner
    # This uses model-specific cost methods for accurate comparison
    baseline = planner.extract_baseline_plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=desires,  # must have current_clusters populated
        extra_model_arguments={"copies_per_region": 3},
    )

    # Compare with custom tolerances
    result = compare_plans(
        baseline,
        recommendation,
        tolerances=ResourceTolerances(
            annual_cost=ignore_resource(),  # Don't care about cost
            cpu=lte(1.05),                 # CPU can be at most 5% over baseline
            disk=plus_or_minus(0.10),      # Storage within ±10%
        ),
    )

    if result.is_equivalent:
        print("Current capacity is sufficient")
    else:
        print("Capacity adjustments needed:")
        for diff in result.get_out_of_tolerance():
            print(f"  - {diff}")
"""

from functools import lru_cache
from itertools import chain

from pydantic import ConfigDict

from service_capacity_modeling.enum_utils import enum_docstrings
from service_capacity_modeling.enum_utils import StrEnum

from service_capacity_modeling.interface import (
    CapacityPlan,
    default_reference_shape,
    ExcludeUnsetModel,
    Instance,
)
from service_capacity_modeling.models.common import get_disk_size_gib


@enum_docstrings
class ResourceType(StrEnum):
    """Resource types that can be compared between capacity plans."""

    annual_cost = "annual_cost"
    """Annual cost in dollars"""

    cpu = "cpu"
    """CPU cores required"""

    mem_gib = "mem_gib"
    """Memory in GiB"""

    disk_gib = "disk_gib"
    """Disk storage in GiB"""

    network_mbps = "network_mbps"
    """Network bandwidth in Mbps"""


# Tolerance for floating-point comparisons (e.g., exact_match)
_FLOAT_TOLERANCE = 1e-9


class Tolerance(ExcludeUnsetModel):
    """Tolerance bounds for comparison / baseline ratio.

    The bounds define the acceptable range for the ratio:
    - ratio > 1.0 = comparison exceeds baseline (need more)
    - ratio < 1.0 = baseline exceeds comparison (have extra capacity)
    - ratio = 1.0 = exact match

    Examples:
        Tolerance(lower=0.0, upper=1.1)            # lte(1.1) - ratio ≤ 1.1
        Tolerance(lower=0.9, upper=float('inf'))   # gte(0.9) - ratio ≥ 0.9
        Tolerance(lower=0.9, upper=1.1)            # tolerance(0.9, 1.1)
    """

    lower: float = 1.0
    """Lower bound for ratio (< 1.0 means can have extra capacity)"""

    upper: float = 1.0
    """Upper bound for ratio (> 1.0 means requirement can exceed baseline)"""

    model_config = ConfigDict(frozen=True)

    def __contains__(self, ratio: float) -> bool:
        """Check if a ratio is within tolerance bounds.

        Args:
            ratio: comparison / baseline

        Returns:
            True if within bounds (equivalent), False otherwise
        """
        return self.lower <= ratio <= self.upper


# -----------------------------------------------------------------------------
# Tolerance helper functions
# -----------------------------------------------------------------------------


@lru_cache(256)
def tolerance(lower: float, upper: float) -> Tolerance:
    """Create a tolerance with explicit ratio bounds.

    Args:
        lower: Lower bound for ratio (e.g., 0.9 means comparison ≥ 0.9× baseline)
        upper: Upper bound for ratio (e.g., 1.1 means comparison ≤ 1.1× baseline)

    Returns:
        Tolerance with specified bounds

    Example:
        >>> t = tolerance(0.9, 1.1)  # Allow 0.9× to 1.1× baseline
        >>> 0.95 in t  # 0.95× baseline: True
        >>> 1.05 in t  # 1.05× baseline: True
        >>> 1.15 in t  # 1.15× baseline: False
    """
    return Tolerance(lower=lower, upper=upper)


def lte(upper: float) -> Tolerance:
    """Ratio must be ≤ upper bound.

    Use this to limit how much the requirement can exceed the baseline.
    Any amount of extra capacity (ratio < 1.0) is acceptable.

    Args:
        upper: Upper bound as a ratio (e.g., 1.1 means comparison ≤ 1.1× baseline).

    Example:
        With baseline=100:

        >>> tol = lte(1.1)  # ratio must be ≤ 1.1

        comparison=110 → ratio = 1.1 → OK (at boundary)
        comparison=111 → ratio = 1.11 → NOT OK
        comparison=90  → ratio = 0.9 → OK (extra capacity)
    """
    return tolerance(0.0, upper)


def gte(lower: float) -> Tolerance:
    """Ratio must be ≥ lower bound.

    Use this to limit how much extra capacity is acceptable.
    Any amount of requirement exceeding baseline (ratio > 1.0) is acceptable.

    Args:
        lower: Lower bound as a ratio (e.g., 0.9 means comparison ≥ 0.9× baseline).

    Example:
        With baseline=100:

        >>> tol = gte(0.9)  # ratio must be ≥ 0.9

        comparison=90  → ratio = 0.9 → OK (at boundary)
        comparison=89  → ratio = 0.89 → NOT OK (too much extra)
        comparison=110 → ratio = 1.1 → OK (need more)
    """
    return tolerance(lower, float("inf"))


def plus_or_minus(percent: float) -> Tolerance:
    """Create a symmetric tolerance of ±percent around 1.0.

    Args:
        percent: The tolerance as a decimal (e.g., 0.10 for ±10%).

    Example:
        >>> tol = plus_or_minus(0.10)  # ratio must be 0.9 to 1.1

        comparison=110, baseline=100 → ratio = 1.1 → OK (at boundary)
        comparison=89, baseline=100 → ratio = 0.89 → NOT OK
    """
    return tolerance(1.0 - percent, 1.0 + percent)


def exact_match() -> Tolerance:
    """Create a tolerance requiring exact match (ratio = 1.0)."""
    return tolerance(1.0 - _FLOAT_TOLERANCE, 1.0 + _FLOAT_TOLERANCE)


def ignore_resource() -> Tolerance:
    """Create a tolerance that ignores a resource (any ratio acceptable)."""
    return tolerance(0.0, float("inf"))


class ResourceTolerances(ExcludeUnsetModel):
    """Per-resource tolerance configuration for plan comparison.

    Set specific tolerances to override for individual resources.
    Unset resources fall back to `default`.

    Example:
        >>> tolerances = ResourceTolerances(
        ...     default=lte(1.1),          # comparison ≤ 1.1× baseline
        ...     cpu=lte(1.05),             # Stricter for CPU
        ...     annual_cost=ignore_resource(),  # Don't care about cost
        ... )
    """

    default: Tolerance = plus_or_minus(0.10)
    """Default tolerance: ±10% (ratio must be 0.9× to 1.1× baseline)"""

    annual_cost: Tolerance | None = None
    """Tolerance for annual cost comparison"""

    cpu: Tolerance | None = None
    """Tolerance for CPU cores comparison"""

    memory: Tolerance | None = None
    """Tolerance for memory (GiB) comparison"""

    disk: Tolerance | None = None
    """Tolerance for disk (GiB) comparison"""

    network: Tolerance | None = None
    """Tolerance for network (Mbps) comparison"""

    def get_tolerance(self, resource: ResourceType) -> Tolerance:
        """Get tolerance for a resource type, falling back to default."""
        match resource:
            case ResourceType.annual_cost:
                return self.annual_cost or self.default
            case ResourceType.cpu:
                return self.cpu or self.default
            case ResourceType.mem_gib:
                return self.memory or self.default
            case ResourceType.disk_gib:
                return self.disk or self.default
            case ResourceType.network_mbps:
                return self.network or self.default
            case _:
                raise ValueError(f"Unknown resource type: {resource}")


class ResourceComparison(ExcludeUnsetModel):
    """Represents a comparison of a resource between baseline and comparison plans.

    Properties:
    - is_equivalent: ratio is within tolerance bounds
    - exceeds_upper_bound: ratio > upper bound (need more than tolerance allows)
    - exceeds_lower_bound: ratio < lower bound (have more extra than tolerance allows)
    """

    resource: ResourceType
    """Resource type being compared"""

    baseline_value: float
    """Value from the current deployment"""

    comparison_value: float
    """Value from the recommendation"""

    tolerance: Tolerance
    """The tolerance bounds that were applied for this comparison"""

    @property
    def ratio(self) -> float:
        """Ratio: comparison / baseline.

        - ratio > 1.0 = comparison exceeds baseline (need more)
        - ratio < 1.0 = baseline exceeds comparison (have extra capacity)
        - ratio = 1.0 = exact match

        Examples:
        - Baseline=100, Comparison=110 → ratio = 1.1 (need 10% more)
        - Baseline=100, Comparison=90 → ratio = 0.9 (have 10% extra)
        """
        if self.baseline_value == 0:
            if self.comparison_value == 0:
                return 1.0  # Both zero = exact match
            return float("inf")  # comparison > 0, baseline = 0
        return self.comparison_value / self.baseline_value

    @property
    def is_equivalent(self) -> bool:
        """True if the ratio is within tolerance bounds."""
        return self.ratio in self.tolerance

    @property
    def exceeds_lower_bound(self) -> bool:
        """True if ratio < lower bound (too much extra capacity)."""
        return self.ratio < self.tolerance.lower

    @property
    def exceeds_upper_bound(self) -> bool:
        """True if ratio > upper bound (requirement exceeds tolerance)."""
        return self.ratio > self.tolerance.upper

    def __str__(self) -> str:
        """Human-readable explanation of the comparison."""
        if self.tolerance.upper == float("inf"):
            bounds_str = f"≥ {self.tolerance.lower:.2f}×"
        elif self.tolerance.lower == 0.0:
            bounds_str = f"≤ {self.tolerance.upper:.2f}×"
        else:
            bounds_str = f"{self.tolerance.lower:.2f}× to {self.tolerance.upper:.2f}×"

        if self.is_equivalent:
            return (
                f"{self.resource.value}: {self.ratio:.2f}× "
                f"(within tolerance: {bounds_str})"
            )
        else:
            bound = "lower" if self.exceeds_lower_bound else "upper"
            return (
                f"{self.resource.value}: exceeds {bound} bound, "
                f"ratio={self.ratio:.2f}× (baseline={self.baseline_value:.2f}, "
                f"comparison={self.comparison_value:.2f}, tolerance: {bounds_str})"
            )


class PlanComparisonResult(ExcludeUnsetModel):
    """Result of comparing two capacity plans for equivalence."""

    is_equivalent: bool
    """True if plans are within tolerance, False if significant differences"""

    comparisons: dict[ResourceType, ResourceComparison] = {}
    """Resource comparisons keyed by resource type"""

    @property
    def cpu(self) -> ResourceComparison:
        """Get CPU comparison result."""
        return self.comparisons[ResourceType.cpu]

    @property
    def memory(self) -> ResourceComparison:
        """Get memory comparison result."""
        return self.comparisons[ResourceType.mem_gib]

    @property
    def disk(self) -> ResourceComparison:
        """Get disk comparison result."""
        return self.comparisons[ResourceType.disk_gib]

    @property
    def network(self) -> ResourceComparison:
        """Get network comparison result."""
        return self.comparisons[ResourceType.network_mbps]

    @property
    def annual_cost(self) -> ResourceComparison:
        """Get annual cost comparison result."""
        return self.comparisons[ResourceType.annual_cost]

    def get_out_of_tolerance(self) -> list[ResourceComparison]:
        """Get only comparisons that exceed tolerance bounds."""
        return [c for c in self.comparisons.values() if not c.is_equivalent]


def to_reference_cores(core_count: float, instance: Instance) -> float:
    """Convert instance cores to reference-equivalent cores.

    This is the inverse of normalize_cores() from models.common. While
    normalize_cores answers "how many target cores to match N reference cores",
    this answers "how many reference cores is N instance cores equivalent to".

    Mathematically equivalent to:
        normalize_cores(core_count, target=default_reference_shape, reference=instance)

    but returns float instead of ceiling int. We need float precision for
    accurate ratio comparisons - ceiling would distort ratios by up to ~3%.

    See normalize_cores() in models/common.py for the original implementation.

    Args:
        core_count: Number of cores on the instance
        instance: The instance shape (with cpu_ghz and cpu_ipc_scale)

    Returns:
        Equivalent cores on default_reference_shape (2.3 GHz, IPC 1.0)

    Example:
        # 32 cores on a 2.4 GHz instance = 33.4 reference cores
        to_reference_cores(32, instance_at_2_4_ghz)  # → 33.39
    """
    instance_speed = instance.cpu_ghz * instance.cpu_ipc_scale
    ref_speed = default_reference_shape.cpu_ghz * default_reference_shape.cpu_ipc_scale
    return core_count * (instance_speed / ref_speed)


def _aggregate_resources(plan: CapacityPlan) -> dict[ResourceType, float]:
    """Aggregate resource values from a plan's candidate_clusters.

    All resources are computed from candidate_clusters (the actual provisioned
    instances), NOT from requirements (the demand). CPU is normalized to
    default_reference_shape using IPC and frequency factors for consistent
    comparison across different instance types.
    """
    totals: dict[ResourceType, float] = {
        ResourceType.cpu: 0.0,
        ResourceType.mem_gib: 0.0,
        ResourceType.disk_gib: 0.0,
        ResourceType.network_mbps: 0.0,
    }

    for cluster in chain(
        plan.candidate_clusters.zonal, plan.candidate_clusters.regional
    ):
        totals[ResourceType.cpu] += to_reference_cores(
            cluster.instance.cpu * cluster.count, cluster.instance
        )
        totals[ResourceType.mem_gib] += cluster.instance.ram_gib * cluster.count
        totals[ResourceType.network_mbps] += cluster.instance.net_mbps * cluster.count
        cluster_drive = cluster.attached_drives[0] if cluster.attached_drives else None
        totals[ResourceType.disk_gib] += (
            get_disk_size_gib(cluster_drive, cluster.instance) * cluster.count
        )

    return totals


def compare_plans(
    baseline: CapacityPlan,
    comparison: CapacityPlan,
    tolerances: ResourceTolerances | None = None,
) -> PlanComparisonResult:
    """Compare two capacity plans to determine if they are roughly equivalent.

    This function compares plans across multiple dimensions (cost, CPU, memory,
    disk, network) and determines if the differences are significant based on
    the provided tolerance bounds.

    Args:
        baseline: The reference plan (e.g., current production deployment)
        comparison: The plan to compare against baseline (e.g., new recommendation)
        tolerances: Per-resource tolerance configuration. If None, uses defaults.

    Returns:
        PlanComparisonResult containing:
        - is_equivalent: True if all resources within tolerance, False otherwise
        - differences: Dict of ResourceDifference for ALL resources (use
          get_out_of_tolerance() to filter to only problematic ones)

    Example:
        >>> result = compare_plans(baseline, recommended)
        >>> if result.cpu.exceeds_lower_bound:
        ...     print("Current has excess CPU capacity")
        >>> for diff in result.get_out_of_tolerance():
        ...     print(diff)  # Human-readable explanation
    """
    if tolerances is None:
        tolerances = ResourceTolerances()

    baseline_resources = _aggregate_resources(baseline)
    comparison_resources = _aggregate_resources(comparison)

    comparisons = {
        rt: ResourceComparison(
            resource=rt,
            baseline_value=b_val,
            comparison_value=comparison_resources[rt],
            tolerance=tolerances.get_tolerance(rt),
        )
        for rt, b_val in baseline_resources.items()
    }
    comparisons[ResourceType.annual_cost] = ResourceComparison(
        resource=ResourceType.annual_cost,
        baseline_value=float(baseline.candidate_clusters.total_annual_cost),
        comparison_value=float(comparison.candidate_clusters.total_annual_cost),
        tolerance=tolerances.get_tolerance(ResourceType.annual_cost),
    )

    return PlanComparisonResult(
        is_equivalent=all(c.is_equivalent for c in comparisons.values()),
        comparisons=comparisons,
    )



"""Tests for plan comparison utility."""

import pytest
from decimal import Decimal

from service_capacity_modeling.interface import (
    CapacityPlan,
    CapacityRequirement,
    Clusters,
    Drive,
    Instance,
    Lifecycle,
    Platform,
    Requirements,
    ZoneClusterCapacity,
    certain_float,
    certain_int,
)
from service_capacity_modeling.models.plan_comparison import (
    compare_plans,
    exact_match,
    gte,
    ignore_resource,
    lte,
    plus_or_minus,
    ResourceComparison,
    ResourceTolerances,
    ResourceType,
    Tolerance,
    tolerance,
    to_reference_cores,
)


# -----------------------------------------------------------------------------
# Test helpers
# -----------------------------------------------------------------------------


def _create_instance(
    *,
    name: str = "test.xlarge",
    cpu: int = 8,
    cpu_ghz: float = 2.3,  # Same as default_reference_shape for 1:1 normalization
    cpu_ipc_scale: float = 1.0,
    ram_gib: float = 32,
    net_mbps: float = 10000,
    drive: Drive | None = None,
) -> Instance:
    """Helper to create a test instance."""
    return Instance(
        name=name,
        cpu=cpu,
        cpu_ghz=cpu_ghz,
        cpu_ipc_scale=cpu_ipc_scale,
        ram_gib=ram_gib,
        net_mbps=net_mbps,
        drive=drive,
        lifecycle=Lifecycle.stable,
        platforms=[Platform.amd64],
    )


def _create_plan(
    *,
    cpu_cores: int = 100,
    mem_gib: float = 200.0,
    disk_gib: float = 1000.0,
    network_mbps: float = 5000.0,
    annual_cost: float = 10000.0,
    count: int = 1,
) -> CapacityPlan:
    """Helper to create test plans with specified resources.

    CPU is computed from candidate_clusters, so cpu_cores sets the instance CPU
    with count=1. The instance uses default_reference_shape IPC/GHz for 1:1
    normalization (cpu_cores directly equals reference-equivalent cores).

    Memory, disk, and network are set on the instance so that cluster-based
    aggregation produces the expected values. Requirements use decoy values
    (9999) so that reading from the wrong source produces obvious failures.
    """
    drive = Drive(name="test-drive", size_gib=int(disk_gib)) if disk_gib else None
    instance = _create_instance(
        cpu=cpu_cores,
        ram_gib=mem_gib,
        net_mbps=network_mbps,
        drive=drive,
    )
    # Decoy requirement values — if _aggregate_resources reads from requirements
    # instead of clusters, tests asserting baseline_value/comparison_value will fail.
    requirement = CapacityRequirement(
        requirement_type="test",
        cpu_cores=certain_int(9999),
        mem_gib=certain_float(9999.0),
        disk_gib=certain_float(9999.0),
        network_mbps=certain_float(9999.0),
    )
    cluster = ZoneClusterCapacity(
        cluster_type="test",
        count=count,
        instance=instance,
        annual_cost=annual_cost,
    )
    return CapacityPlan(
        requirements=Requirements(zonal=[requirement]),
        candidate_clusters=Clusters(
            annual_costs={"test": Decimal(str(annual_cost))},
            zonal=[cluster],
        ),
    )


# -----------------------------------------------------------------------------
# to_reference_cores tests
# -----------------------------------------------------------------------------


class TestToReferenceCores:
    """Tests for to_reference_cores() utility function."""

    def test_same_as_reference_shape(self):
        """Instance matching reference shape returns cores unchanged.

        default_reference_shape: 2.3 GHz, IPC 1.0
        100 cores @ 2.3 GHz, IPC 1.0 → 100 × (2.3×1.0)/(2.3×1.0) = 100
        """
        instance = _create_instance(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=1.0)
        assert to_reference_cores(100, instance) == 100.0

    def test_faster_instance_gives_more_reference_cores(self):
        """Faster instance (higher GHz) yields more reference-equivalent cores.

        100 cores @ 3.0 GHz, IPC 1.0 → 100 × (3.0×1.0)/(2.3×1.0) = 130.4
        """
        instance = _create_instance(cpu=100, cpu_ghz=3.0, cpu_ipc_scale=1.0)
        result = to_reference_cores(100, instance)
        assert 130.0 < result < 131.0  # 130.43

    def test_higher_ipc_gives_more_reference_cores(self):
        """Higher IPC yields more reference-equivalent cores.

        100 cores @ 2.3 GHz, IPC 2.0 → 100 × (2.3×2.0)/(2.3×1.0) = 200
        """
        instance = _create_instance(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=2.0)
        assert to_reference_cores(100, instance) == 200.0

    def test_slower_instance_gives_fewer_reference_cores(self):
        """Slower instance (lower GHz or IPC) yields fewer reference-equivalent cores.

        100 cores @ 2.0 GHz, IPC 0.8 → 100 × (2.0×0.8)/(2.3×1.0) = 69.6
        """
        instance = _create_instance(cpu=100, cpu_ghz=2.0, cpu_ipc_scale=0.8)
        result = to_reference_cores(100, instance)
        assert 69.0 < result < 70.0  # 69.57

    def test_returns_float_not_int(self):
        """Returns float for precise ratio calculations (unlike normalize_cores).

        32 cores @ 2.4 GHz, IPC 1.0 → 32 × (2.4×1.0)/(2.3×1.0) = 33.39
        normalize_cores would return ceil(33.39) = 34, but we need 33.39
        """
        instance = _create_instance(cpu=32, cpu_ghz=2.4, cpu_ipc_scale=1.0)
        result = to_reference_cores(32, instance)
        assert isinstance(result, float)
        assert 33.3 < result < 33.5  # 33.39, not 34


# -----------------------------------------------------------------------------
# Tolerance tests
# -----------------------------------------------------------------------------


class TestTolerance:
    def test_boundaries_are_inclusive(self):
        t = Tolerance(lower=0.9, upper=1.2)
        assert 0.9 in t
        assert 1.2 in t
        assert 0.89 not in t
        assert 1.21 not in t


class TestToleranceHelpers:
    """Tests for tolerance helper functions."""

    def test_tolerance(self):
        t = tolerance(0.95, 1.20)
        assert t.lower == 0.95
        assert t.upper == 1.20

    def test_plus_or_minus(self):
        t = plus_or_minus(0.10)
        assert t.lower == 0.9
        assert t.upper == 1.1

    def test_lte(self):
        t = lte(1.1)
        assert t.lower == 0.0
        assert t.upper == 1.1
        assert 0.5 in t
        assert 1.2 not in t

    def test_gte(self):
        t = gte(0.9)
        assert t.lower == 0.9
        assert t.upper == float("inf")
        assert 2.0 in t
        assert 0.8 not in t

    def test_exact_match(self):
        t = exact_match()
        assert 1.0 in t
        assert 0.99 not in t

    def test_ignore_resource(self):
        t = ignore_resource()
        assert 0.01 in t
        assert 100.0 in t


# -----------------------------------------------------------------------------
# ResourceTolerances tests
# -----------------------------------------------------------------------------


class TestResourceTolerances:
    """Tests for per-resource tolerance configuration."""

    def test_fallback_to_default(self):
        """get_tolerance falls back to default when not specified."""
        tolerances = ResourceTolerances(
            default=plus_or_minus(0.10),
            cpu=plus_or_minus(0.05),
        )
        # CPU has specific tolerance
        assert tolerances.get_tolerance(ResourceType.cpu).lower == 0.95
        # Memory falls back to default
        assert tolerances.get_tolerance(ResourceType.mem_gib).lower == 0.9


# -----------------------------------------------------------------------------
# ResourceComparison tests
# -----------------------------------------------------------------------------


class TestResourceComparison:
    """Tests for ResourceComparison ratio calculation and direction detection."""

    def test_ratio_calculation(self):
        """Ratio is calculated as comparison / baseline."""
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=100,
            comparison_value=75,
            tolerance=lte(1.1),
        )
        assert comp.ratio == pytest.approx(0.75)
        assert comp.baseline_value == 100
        assert comp.comparison_value == 75

    def test_ratio_zero_baseline_nonzero_comparison(self):
        """Zero baseline with non-zero comparison returns infinity."""
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=0,
            comparison_value=100,
            tolerance=lte(1.1),
        )
        assert comp.ratio == float("inf")
        assert comp.exceeds_upper_bound

    def test_ratio_both_zero(self):
        """Both zero returns ratio of 1.0 (exact match)."""
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=0,
            comparison_value=0,
            tolerance=lte(1.1),
        )
        assert comp.ratio == 1.0
        assert comp.is_equivalent

    @pytest.mark.parametrize(
        "baseline,comparison,expected_bound",
        [
            (100, 120, "upper"),  # 1.2 > 1.1 upper bound
            (100, 130, "upper"),  # 1.3 > 1.1 upper bound
            (100, 50, "lower"),  # 0.5 < 0.9 lower bound (with symmetric)
        ],
    )
    def test_exceeds_bound_detection(self, baseline, comparison, expected_bound):
        tol = plus_or_minus(0.10) if expected_bound == "lower" else lte(1.1)
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=baseline,
            comparison_value=comparison,
            tolerance=tol,
        )
        if expected_bound == "upper":
            assert comp.exceeds_upper_bound
            assert not comp.exceeds_lower_bound
        else:
            assert comp.exceeds_lower_bound
            assert not comp.exceeds_upper_bound

    def test_is_equivalent_when_within_tolerance(self):
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=100,
            comparison_value=105,
            tolerance=lte(1.1),
        )
        assert comp.is_equivalent
        assert not comp.exceeds_upper_bound
        assert not comp.exceeds_lower_bound

    def test_str_shows_bound_exceeded(self):
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=100,
            comparison_value=130,
            tolerance=plus_or_minus(0.10),
        )
        assert "exceeds upper bound" in str(comp)

        comp_lower = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=100,
            comparison_value=50,
            tolerance=plus_or_minus(0.10),
        )
        assert "exceeds lower bound" in str(comp_lower)

    def test_str_shows_within_tolerance(self):
        """__str__ indicates when within tolerance."""
        comp = ResourceComparison(
            resource=ResourceType.cpu,
            baseline_value=100,
            comparison_value=105,
            tolerance=lte(1.1),
        )
        assert "within tolerance" in str(comp)


# -----------------------------------------------------------------------------
# compare_plans tests
# -----------------------------------------------------------------------------


class TestComparePlans:
    """Tests for compare_plans function."""

    @pytest.mark.parametrize("count", [1, 3])
    def test_identical_plans_are_equivalent(self, count):
        """Identical plans should be equivalent.

        Also verifies that resource values are aggregated from candidate_clusters
        (not requirements, which use decoy value 9999), and scaled by count.
        """
        plan = _create_plan(count=count)
        result = compare_plans(plan, plan)
        assert result.is_equivalent
        assert all(c.is_equivalent for c in result.comparisons.values())
        # Verify values come from clusters (scaled by count), not decoy reqs (9999)
        assert result.cpu.baseline_value == 100.0 * count
        assert result.memory.baseline_value == 200.0 * count
        assert result.disk.baseline_value == 1000.0 * count
        assert result.network.baseline_value == 5000.0 * count
        assert result.annual_cost.baseline_value == 10000.0

    def test_within_tolerance_is_equivalent(self):
        """Plans within default tolerance are equivalent."""
        baseline = _create_plan(cpu_cores=100)
        comparison = _create_plan(cpu_cores=105)  # ratio = 1.05 < 1.1
        result = compare_plans(baseline, comparison)
        assert result.is_equivalent

    @pytest.mark.parametrize("count", [1, 3])
    def test_exceeding_tolerance_is_not_equivalent(self, count):
        baseline = _create_plan(cpu_cores=100, count=count)
        comparison = _create_plan(cpu_cores=120, count=count)
        result = compare_plans(baseline, comparison)
        assert not result.is_equivalent
        assert result.cpu.exceeds_upper_bound
        assert result.cpu.baseline_value == 100.0 * count
        assert result.cpu.comparison_value == 120.0 * count

    @pytest.mark.parametrize(
        "resource_kwarg,accessor",
        [
            ("cpu_cores", "cpu"),
            ("mem_gib", "memory"),
            ("disk_gib", "disk"),
            ("network_mbps", "network"),
            ("annual_cost", "annual_cost"),
        ],
    )
    def test_exceeds_upper_bound_for_each_resource(self, resource_kwarg, accessor):
        baseline_kwargs = {
            "cpu_cores": 100,
            "mem_gib": 200.0,
            "disk_gib": 1000.0,
            "network_mbps": 5000.0,
            "annual_cost": 10000.0,
        }
        comparison_kwargs = baseline_kwargs.copy()
        comparison_kwargs[resource_kwarg] = baseline_kwargs[resource_kwarg] * 1.3

        baseline = _create_plan(**baseline_kwargs)
        comparison = _create_plan(**comparison_kwargs)
        result = compare_plans(baseline, comparison)

        resource_result = getattr(result, accessor)
        assert resource_result.exceeds_upper_bound
        assert not resource_result.is_equivalent

    def test_exceeds_lower_bound_with_symmetric_tolerance(self):
        baseline = _create_plan(cpu_cores=100)
        comparison = _create_plan(cpu_cores=50)
        result = compare_plans(
            baseline,
            comparison,
            tolerances=ResourceTolerances(default=plus_or_minus(0.10)),
        )
        assert not result.is_equivalent
        assert result.cpu.exceeds_lower_bound

    def test_get_out_of_tolerance(self):
        """get_out_of_tolerance returns only problematic resources."""
        baseline = _create_plan(cpu_cores=100, mem_gib=200)
        comparison = _create_plan(cpu_cores=120, mem_gib=200)  # Only CPU exceeds
        result = compare_plans(baseline, comparison)

        out_of_tolerance = result.get_out_of_tolerance()
        assert len(out_of_tolerance) == 1
        assert out_of_tolerance[0].resource == ResourceType.cpu

    def test_returns_all_resource_types(self):
        """compare_plans returns comparisons for all 5 resource types."""
        result = compare_plans(_create_plan(), _create_plan())
        assert len(result.comparisons) == 5
        assert set(result.comparisons.keys()) == {
            ResourceType.annual_cost,
            ResourceType.cpu,
            ResourceType.mem_gib,
            ResourceType.disk_gib,
            ResourceType.network_mbps,
        }

    def test_per_resource_tolerances(self):
        baseline = _create_plan(cpu_cores=100, mem_gib=200)
        comparison = _create_plan(cpu_cores=85, mem_gib=170)

        tolerances = ResourceTolerances(
            cpu=plus_or_minus(0.20),
            memory=plus_or_minus(0.10),
        )
        result = compare_plans(baseline, comparison, tolerances=tolerances)

        assert result.cpu.is_equivalent  # 0.85 >= 0.8
        assert not result.memory.is_equivalent  # 0.85 < 0.9
        assert result.memory.exceeds_lower_bound

    def test_cpu_normalization_across_different_ipc(self):
        """CPU comparison normalizes for both GHz and IPC differences.

        Normalization formula:
            ref_cores = cores × (ghz × ipc) / (2.3 × 1.0)

        Where 2.3 GHz and IPC 1.0 is the default_reference_shape.

        Test cases:
        1. Different GHz, same IPC:
           - A: 100 cores @ 2.3 GHz, IPC 1.0 → 100 × 2.3/2.3 = 100 ref
           - B: 77 cores @ 3.0 GHz, IPC 1.0 → 77 × 3.0/2.3 = 100.4 ref
           Ratio ≈ 1.0 → equivalent

        2. Same GHz, different IPC (higher IPC = fewer cores needed):
           - A: 100 cores @ 2.3 GHz, IPC 1.0 → 100 ref
           - B: 50 cores @ 2.3 GHz, IPC 2.0 → 50 × 4.6/2.3 = 100 ref
           Ratio = 1.0 → equivalent

        3. Combined GHz + IPC difference:
           - A: 100 cores @ 2.3 GHz, IPC 1.0 → 100 ref
           - B: 40 cores @ 3.0 GHz, IPC 2.0 → 40 × 6.0/2.3 = 104.3 ref
           Ratio ≈ 1.04 → equivalent (within ±10%)

        4. IPC difference exceeds tolerance (failure case):
           - A: 100 cores @ 2.3 GHz, IPC 1.0 → 100 ref
           - B: 100 cores @ 2.3 GHz, IPC 0.5 → 100 × 1.15/2.3 = 50 ref
           Ratio = 0.5 → under-provisioned (outside ±10%)
        """

        def make_plan(cpu: int, cpu_ghz: float, cpu_ipc_scale: float) -> CapacityPlan:
            inst = _create_instance(
                cpu=cpu, cpu_ghz=cpu_ghz, cpu_ipc_scale=cpu_ipc_scale
            )
            return CapacityPlan(
                requirements=Requirements(
                    zonal=[
                        CapacityRequirement(
                            requirement_type="t", cpu_cores=certain_int(cpu)
                        )
                    ]
                ),
                candidate_clusters=Clusters(
                    annual_costs={"t": Decimal("1000")},
                    zonal=[
                        ZoneClusterCapacity(
                            cluster_type="t", count=1, instance=inst, annual_cost=1000.0
                        )
                    ],
                ),
            )

        # Case 1: Different GHz, same IPC
        baseline = make_plan(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=1.0)
        comparison = make_plan(cpu=77, cpu_ghz=3.0, cpu_ipc_scale=1.0)
        result = compare_plans(baseline, comparison)
        assert result.cpu.is_equivalent
        assert 0.99 < result.cpu.ratio < 1.05  # 100.4/100 ≈ 1.004

        # Case 2: Same GHz, different IPC (2× IPC = half the cores)
        baseline = make_plan(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=1.0)
        comparison = make_plan(cpu=50, cpu_ghz=2.3, cpu_ipc_scale=2.0)
        result = compare_plans(baseline, comparison)
        assert result.cpu.is_equivalent
        assert 0.99 < result.cpu.ratio < 1.01  # Exactly 1.0

        # Case 3: Combined GHz + IPC
        baseline = make_plan(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=1.0)
        comparison = make_plan(cpu=40, cpu_ghz=3.0, cpu_ipc_scale=2.0)
        result = compare_plans(baseline, comparison)
        assert result.cpu.is_equivalent
        assert 1.0 < result.cpu.ratio < 1.1  # 104.3/100 ≈ 1.043

        # Case 4: IPC difference exceeds tolerance (NOT equivalent)
        baseline = make_plan(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=1.0)
        comparison = make_plan(cpu=100, cpu_ghz=2.3, cpu_ipc_scale=0.5)
        result = compare_plans(baseline, comparison)
        assert not result.cpu.is_equivalent
        assert result.cpu.exceeds_lower_bound
        assert 0.49 < result.cpu.ratio < 0.51  # Exactly 0.5



#!/usr/bin/python3
import subprocess
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from service_capacity_modeling.tools.instance_families import INSTANCE_TYPES

print("Loaded instance family count =", len(INSTANCE_TYPES))


def get_auto_shape_path() -> Path:
    """Get the path to auto_shape.py script"""
    current_dir = Path(__file__).parent
    return current_dir / "auto_shape.py"


def build_command(family: str, params: Dict[str, Any], output_path: Path) -> List[str]:
    """Build the command to run auto_shape.py with the appropriate parameters"""
    auto_shape_path = get_auto_shape_path()
    cmd = [sys.executable, str(auto_shape_path)]

    # Add optional parameters if they are provided
    if params.get("xl_iops") is not None:
        cmd.extend(["--xl-iops", params["xl_iops"]])

    if params.get("io_latency_curve") is not None:
        cmd.extend(["--io-latency-curve", params["io_latency_curve"]])

    if params.get("cpu_ipc_scale") is not None:
        cpu_ipc_scale = float(params["cpu_ipc_scale"])
        rounded_cpu_ipc_scale = float(f"{cpu_ipc_scale:.2f}")
        cmd.extend(["--cpu-ipc-scale", str(rounded_cpu_ipc_scale)])

    # Add output path
    cmd.extend(["--output-path", str(output_path)])

    # Add family as positional argument
    cmd.append(family)

    return cmd


def main(debug: bool = True, execute: bool = False, force: bool = False) -> None:
    expected_path = (
        Path(__file__).resolve().parent.parent / "hardware/profiles/shapes/aws"
    )

    print("Checking for files in", expected_path)

    # Check which families need to be generated
    missing_families = {
        family: params
        for family, params in INSTANCE_TYPES.items()
        if force or not (expected_path / f"auto_{family}.json").exists()
    }

    if not missing_families:
        print("All instance family files are present.")
        return

    # Build commands for missing families
    commands = {
        family: build_command(family, params, expected_path)
        for family, params in missing_families.items()
    }

    for family, cmd in commands.items():
        if debug:
            print(f"\nWould run for {family}:\n{' '.join(cmd)}")

        if execute:
            print(f"\nGenerating {family}...")
            try:
                result = subprocess.run(cmd, capture_output=True, text=True, check=True)
                print(f"Successfully generated {family}")
                if debug:
                    print(result.stdout)
            except subprocess.CalledProcessError as e:
                print(f"Error generating {family}: {e}")
                print(e.stderr)

    if not debug and not execute:
        print("\nRun with --execute to generate the missing files")


if __name__ == "__main__":
    # Parse arguments to determine whether to execute commands
    execute_mode = "--execute" in sys.argv
    debug_mode = "--debug" in sys.argv or not execute_mode
    force_mode = "--force" in sys.argv

    main(debug=debug_mode, execute=execute_mode, force=force_mode)



import math
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRegretParameters
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import ClusterCapacity
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionClusterCapacity
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models import CostAwareModel
from service_capacity_modeling.models.common import cluster_infra_cost
from service_capacity_modeling.models.common import compute_stateless_region
from service_capacity_modeling.models.common import network_services
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores


def _estimate_java_app_requirement(
    instance: Instance,
    desires: CapacityDesires,
    failover: bool = True,
    jvm_memory_overhead: float = 1.2,
) -> CapacityRequirement:
    needed_cores = sqrt_staffed_cores(desires)
    needed_network_mbps = simple_network_mbps(desires)

    if failover:
        # For failover provision at 40% utilization
        needed_cores = int(math.ceil(needed_cores * (1 / 0.4)))
        needed_network_mbps = int(math.ceil(needed_network_mbps * (1 / 0.4)))

    # Adjust cores to clock frequency differences against measurement
    needed_cores = normalize_cores(
        core_count=needed_cores,
        target_shape=instance,
        reference_shape=desires.reference_shape,
    )

    # Assume a Java application that can allocate about 1 GiB/s to heap
    # per 2 GiB of heap with some amount of overhead on the network traffic.
    # So for example if we have 512 MiB of network traffic there is some
    # overhead associated with that...

    # TODO: we should probably have the memory bandwidth attached to
    # the instance type, e.g. Intel CPUs and AMD CPUs have different
    # per core memory bandwidth.
    mem_allocation_mbps = needed_network_mbps * jvm_memory_overhead
    heap_allocation_gibps = (mem_allocation_mbps / 8) / 1024
    network_heap = heap_allocation_gibps * 2

    needed_memory_gib = network_heap

    return CapacityRequirement(
        requirement_type="java-app",
        reference_shape=desires.reference_shape,
        cpu_cores=certain_int(needed_cores),
        mem_gib=certain_float(needed_memory_gib),
        network_mbps=certain_float(needed_network_mbps),
        context={
            "network_heap_gib": network_heap,
            "reserved_mem": desires.data_shape.reserved_instance_app_mem_gib,
        },
    )


def _estimate_java_app_region(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    context: RegionContext,
    root_disk_gib: int = 10,
    failover: bool = True,
    jvm_memory_overhead: float = 2,
) -> Optional[CapacityPlan]:
    if drive.name != "gp2":
        return None

    zones_per_region = context.zones_in_region
    requirement = _estimate_java_app_requirement(
        instance, desires, failover, jvm_memory_overhead
    )

    drive = drive.model_copy()
    drive.size_gib = root_disk_gib
    attached_drives = (drive,)

    cluster: RegionClusterCapacity = compute_stateless_region(
        instance=instance,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_memory_gib=requirement.mem_gib.mid,
        needed_network_mbps=requirement.network_mbps.mid,
        num_zones=zones_per_region,
    )
    cluster.cluster_type = NflxJavaAppCapacityModel.cluster_type
    cluster.attached_drives = attached_drives

    # Generally don't want giant clusters
    # Especially not above 1000 because some load balancers struggle
    # with such large clusters

    if cluster.count <= 256:
        costs = NflxJavaAppCapacityModel.cluster_costs(
            service_type=NflxJavaAppCapacityModel.service_name,
            regional_clusters=[cluster],
        )
        services = NflxJavaAppCapacityModel.service_costs(
            service_type=NflxJavaAppCapacityModel.service_name,
            context=context,
            desires=desires,
            extra_model_arguments={},
        )
        for s in services:
            costs[s.service_type] = s.annual_cost

        return CapacityPlan(
            requirements=Requirements(regional=[requirement]),
            candidate_clusters=Clusters(
                annual_costs=costs,
                regional=[cluster],
                zonal=[],
                services=services,
            ),
        )
    return None


class NflxJavaAppArguments(BaseModel):
    failover: bool = Field(
        default=True, description="If this app participates in failover"
    )
    jvm_memory_overhead: float = Field(
        default=1.2,
        description="How much overhead does the heap have per read byte",
    )
    root_disk_gib: int = Field(
        default=10, description="How many GiB of root volume to attach"
    )


class NflxJavaAppCapacityModel(CapacityModel, CostAwareModel):
    service_name = "nflx-java-app"
    cluster_type = "nflx-java-app"

    @staticmethod
    def cluster_costs(
        service_type: str,
        zonal_clusters: Sequence[ClusterCapacity] = (),
        regional_clusters: Sequence[ClusterCapacity] = (),
    ) -> Dict[str, float]:
        return cluster_infra_cost(
            service_type,
            zonal_clusters,
            regional_clusters,
            cluster_type=NflxJavaAppCapacityModel.cluster_type,
        )

    @staticmethod
    def service_costs(
        service_type: str,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> List[ServiceCapacity]:
        # TODO(matthewho): Currently returns empty because RegionContext is
        # created without services. Need to determine if stateless apps should
        # have cross-zone costs (copies_per_region=2 implies 1 cross-AZ hop).
        _ = (context, extra_model_arguments)
        return network_services(
            service_type,
            RegionContext(num_regions=1),
            desires,
            copies_per_region=2,
        )

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        failover: bool = extra_model_arguments.get("failover", True)
        jvm_memory_overhead: float = extra_model_arguments.get(
            "jvm_memory_overhead", 1.2
        )
        root_disk_gib: int = extra_model_arguments.get("root_disk_gib", 10)

        return _estimate_java_app_region(
            instance=instance,
            drive=drive,
            desires=desires,
            failover=failover,
            root_disk_gib=root_disk_gib,
            jvm_memory_overhead=jvm_memory_overhead,
            context=context,
        )

    @staticmethod
    def description() -> str:
        return "Netflix Streaming Java App Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxJavaAppArguments.model_json_schema()

    @staticmethod
    def regret(
        regret_params: CapacityRegretParameters,
        optimal_plan: CapacityPlan,
        proposed_plan: CapacityPlan,
    ) -> Dict[str, float]:
        regret = super(NflxJavaAppCapacityModel, NflxJavaAppCapacityModel).regret(
            regret_params, optimal_plan, proposed_plan
        )
        regret["disk_space"] = 0
        return regret

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.never,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=2, confidence=0.98
                    ),
                    # "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.5,
                        maximum_value=10,
                        low=1,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        low=1, mid=2, high=5, confidence=0.98
                    ),
                ),
                data_shape=DataShape(
                    # Assume 4 GiB heaps
                    reserved_instance_app_mem_gib=4
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.never,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=128, high=1024, confidence=0.95
                    ),
                    # Throughput ops can be slower
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=4, high=8, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=1, high=5, confidence=0.98
                    ),
                    # "Tens of millisecond SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.5,
                        maximum_value=100,
                        low=1,
                        mid=5,
                        high=40,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.5,
                        maximum_value=100,
                        low=1,
                        mid=5,
                        high=40,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    # Assume 4 GiB heaps
                    reserved_instance_app_mem_gib=4
                ),
            )


nflx_java_app_capacity_model = NflxJavaAppCapacityModel()
